﻿:Namespace SALTUtils

⍝   Utility Functions used by SALT

   ⍝ Version 2.31, 2012/9/10 help by folder, <GetSpiceList> to include unew and rebuild, removed V11 ref in specialName

    ⎕io←1 ⋄ ⎕wx←⎕ml←3 ⋄ CR←⎕av[4] ⋄ ⎕USING←0⍴⊂'' ⋄ MONITOR←DEBUG←0 ⋄ MONITORNAME←'#.UCMDMonitor'

    (WIN WWIN)←∨\(FS←'.'⎕wg 'aplversion')[3]∊⍨(,'W')'WINE'
    ⎕fx'r←(f Ver)n' ('r←',(4↑2⊃FS),'f n')
    ⎕fx'r←RSI n'((1+<Ver 13)⊃'r←(n+1)⊃⎕RSI' 'r←⍎(n+1)⊃⎕NSI')
    FS←'/\'[1+WIN]
    LINDEL←(~WIN)↓13 10
    AZaz←' ',⎕A,[0.1]'abcdefghijklmnopqrstuvwxyz' ⍝ used for sorting caseless

⍝ ===== SPICE/UCMD related code

    UVersion←1.28 ⍝ User Commands version

    Name Group Desc Parse ObjName FullName←⍳6 ⍝ column indices for the UCMD list

    ∇ r←SpiceHELP
      r←1↓⊃,/CR,¨2↓¨2↓⎕NR'SpiceHELP'
⍝
⍝   User Commands - General Help
⍝
⍝ A User Command (UCMD) consists of a single unique keyword mapped to some code written in a
⍝ class or namespace in a SALT file (a Unicode text file). UCMDs used to be called Spice commands.
⍝
⍝ Under V12.1 and above, UCMDs can be entered in the session by preceding them with ],
⍝ e.g. command 'ABC' would be issued as
⍝      ]ABC
⍝ Below V12.1 User Commands can be entered in the Spice input area if enabled.
⍝
⍝ UCMD files are located in specific folders. After installation all public UCMD files
⍝ will be located under the SALT\Spice folder. By default this is where commands will be looked for.
⍝ This folder is reported by the SALT command 'settings' under 'cmddir' and can be altered if
⍝ desired. For example, if you wish to add your own command folder of User Commands you should
⍝ tell SALT like this:
⍝       ]settings  cmddir ,\my\own\folder       (or in the Spice input area)
⍝ or
⍝       ⎕SE.SALT.Settings 'cmddir ,\my\own\folder'   (with SALT)
⍝ where the comma means 'add to the existing setting'. Without the comma, only your commands
⍝ and the SALT commands will be available.
⍝
⍝ To get a list of all the commands found in the current command folder enter ']?'.
⍝ To get a one liner help for each command found in the current command folder enter ']?+'.
⍝ To get a one liner help for each command found in a specific folder enter ']? \folder'.
⍝ Either way will show the commands found in all the Spice files found along with a short description.
⍝ To get help on a specific command enter ']? commandname' or ]?? commandname'. The later should
⍝ provide more detailed help. See  ]?save  and  ]??save  for an example.
⍝
⍝ To get help on the LAST APL ERROR that occurred, use ]DMX
⍝
    ∇

    ∇ com←ln displayNames(arg list);c;ll;cs;i;t;dn;L;ord
    ⍝ List all the cmds available. arg is ]?'s arg, list is the list of all UCMDs, ln is the lowercase version of names
      com←0 ⍝ result if nothing to show
      :If 0∊⍴arg
          ll←CR,∊CR,¨(⎕PW-1)BJp'Type "]?+" for a summary or "]??" for general help or "]?CMD" for info on command CMD.'
          :If 0∊+/';'=t←CMDDIR   ⍝ only 1 folder? must be SALT's
              com←CR,(0⍕⍴ln),' commands:'
              com←ll,⍨com,CR,,CR,⎕PW 8 WidthFit{⍵[AZaz⍋⊃⍵]}ln
          :Else ⍝ multiple paths, show cmds for each - this is a bit more work:
              dn←⍒∊L←⍴¨c←(1↓¨{⍵⊂⍨';'=⍵}';',t),⊂BootPath'Spice' ⋄ com←'' ⍝ we search by path length to cover the case \X,\X\Y
              t←c[dn]∘{1⍳⍨⍺≡¨L[dn]↑¨⊂⍵}¨list[;FullName] ⍝ INDEXERR if no matching folder for a cmd
              :For i :In ⍳⍴c{⍺}ord←dn[t]
                  :If ∨/cs←i=ord
                      com,←∊CR,¨(⎕PW-1)BJp(0⍕+/cs),' commands in "',(i⊃c),'":',CR
                      com,←CR,⍨,CR,⎕PW 8 WidthFit{⍵[AZaz⍋⊃⍵]}cs⌿ln
                  :EndIf
              :EndFor
              com←com,ll
          :EndIf
      :ElseIf arg≡,'+'           ⍝ display 1 line summary per cmd
          com←helpfor list
      :ElseIf ~isRelPath arg     ⍝ is this a request for help on a path?
          com←arg helpfor 2⊃GetSpiceList arg'*'
      :EndIf
    ∇

    ∇ {R}←{rgn}Spice msg;cmd;arg;list;i;c;find;lNames;rules;nargs;switches;syntax;splitRule;name;rez;helpfor;help;com;findall;spc;reset;cs;details;names;mon;t;DMX
⍝ Called by the UCMD function. The function can also be used directly with a string as argument.
      :If 0∊⎕NC'rgn'
          rgn←1
      :EndIf   ⍝ is the result used?
      R←¯1 ⋄ cmd←msg
     
⍝ Put a stop before the call if requested
      c←×DEBUG∧'-'=¯1↑cmd←rtb cmd
      ((c/go),go~⍨⎕STOP'Spice')⎕STOP'Spice' ⋄ cmd↓⍨←-c ⍝ don't change previous stops, if any
      :If 0<i←+/∧\'?'=cmd←'] 'rlb cmd
          cmd←cmd[⍳i],' ',rlb i↓cmd
      :EndIf ⍝ accept ?cmd
      :If (i>1)∧i=¯1+⍴cmd        ⍝ ]?? ≡ ]help
      :OrIf 'help'≡lCase rtb cmd
          R←SpiceHELP ⋄ →0    ⍝ more than 1 "?" alone or 'help': general help
      :EndIf
      R←0
      (cmd arg)←cmd splitOn1st' ' ⋄ cmd←lCase cmd
      :If 'help'≡cmd
          cmd←,'?'
      :EndIf ⍝ accept "help xxx" instead of "?xxx"
     
     ⍝ In 13.1 ⎕DMX is introduced and used in some ucmds (e.g. ]DMX)
     ⍝ Since Spice may reset it we take a copy here
      DMX←{2::0 ⋄ ⍎'⎕DMX'}0
     
⍝ It is always possible to call Spice directly thru this fn. If Spice's input area has not been
⍝ enabled it will fail unless we automatically boot Spice:
      :If 0=⎕NC'⎕SE.cbbot.bandsalt'
          BootSpice
      :EndIf
     
      :If reset←'ureset'≡cmd            ⍝ allow 'ureset' to force reset
          CMDDIR←{6::0 ⋄ CMDDIR}0
          t←⎕SI∊⍨⊂'GetSpiceList'        ⍝ is this a recursive call thru ⎕FIX?
      :OrIf t<CMDDIR≢⎕SE.SALT.Settings'cmddir'  ⍝ has the Spice folder changed?
          R←⍕1↑⍴⎕IO⊃⌽(CMDDIR ⎕SE.cbbot.bandsalt.Spice.List)←(~rgn)GetSpiceList'*'
          R,←' commands reloaded'
          →reset⍴0
          R←0
      :EndIf
      list←⎕SE.cbbot.bandsalt.Spice.List
      lNames←lCase¨list[;Name Group]
      findall←{b/⍳⍴b←((⍴⍵)↑¨⍺)∊⊂⍵}
      find←{lNames[;1]strIndex⊂⍵}
     
⍝ This function splits the number of arguments from the switches definition and reformats nicely
      splitRule←{' '∧.=R←rlb ⍵:'' ⋄ s←~a←∧\R∊⎕D,l←'sSlL[] ' ⋄ (l{(⍵~⍺),(∨/2 2⍴'sSlL'∊⍵)/'SL'}a/R)(s/R)}
     
      helpfor←{0∊⍴⍵:'No files found in <',⍺,'>'
          c←(,[0.5]∘'='¨'Group' 'Name' 'Description')⍪{⍵[AZaz⍋⍕⍵[;1 2];]}⍵[;Group Name Desc]
          c[1+{⍵/⍳⍴⍵}2≡/c[;1];1]←⊂'' ⍝ blank out all but 1st group names
          '"]??" for general help, "]?CMD" for more specific info on command CMD',⍨CR,⍨,CR,⍨⍕c}
     
⍝ One line help
      :If help←'?'=1↑com←cmd
          →0/⍨0≢R←lNames[;1]displayNames arg list
          details←¯1+'?'+.=com
          com←lCase 1⊃arg splitOn' ' ⍝ the 1st word after '?' is the command to get help on
      :EndIf
     
⍝ We need to bring in the command if it is unknown and AUTOmatic search is ON
      :If 0=i←find com  ⍝ look for exact name
      :AndIf ~'*'∊cmd   ⍝ filter used?
      :AndIf 'auto'≡lCase ⎕SE.SALT.Settings'newcmd'
          (CMDDIR list)←(~rgn)GetSpiceList'*' ⍝ we need to reload everything because the command may not have the same name as the file
          ⎕SE.cbbot.bandsalt.Spice.List←list
          lNames←lCase¨list[;Name Group]
          i←find com ⍝ try again
      :EndIf
     
⍝ If the command is still not found and help is wanted we look into groups and similar names (use of *)
      :If 0=i
      :AndIf (~help)∨0∊⍴i←lNames[;2-'*'∊com]findall com~'*'   ⍝ group name?
          R←⎕TC[1],'* Command "',com,'" not found. Type ]? for a list of all commands.' ⋄ →0
      :EndIf
     
      :If help ⍝ Special case '?'
          :If 1<⍴,i
              R←helpfor list[,i;] ⋄ →0
          :EndIf
          name←Name⊃,list[i;]
          :Trap DEBUG↓0
⍝ If parsing rules are given display them here
              syntax←''
              :If ''≢rules←splitRule Parse⊃,list[i;]
                  (nargs switches)←rules
                  c←⍴syntax←' Syntax: ',(''≢nargs)/nargs,' arguments;'
⍝ Try to put the switches on the same line if small enough:
                  :If ⎕PW≥c+10+(⍴name)+⍴t←(''≢switches)/' accepts switches ',switches
                      syntax,←t
                  :Else ⍝ won't fit, split switches' display
                      t←(+\t=1↑t)⊂t←switches
                      syntax,←' accepts switches ',,⊃,/CR,¨⍕¨(⎕PW-9)BJp t
                  :EndIf
              :EndIf
              com←'Command "',name,'".',syntax
              c←FullName⊃,list[i;]
              spc←⎕SE.SALT.Load'"',c,'" -noname' ⍝ grab the cmd space
              :Trap 2  ⍝ try the detailed syntax
                  com,←CR,,CR,details spc.Help name
              :Else    ⍝ OK, the "old" way then
                  com,←CR,,CR,spc.Help name
              :EndTrap
              com,←(CR=¯1↑com)↓CR,CR,'Script location: ',c,CR
              R←com
          :Else
              R←⎕TC[1],'* Unable to produce Help for command "',name,'":',{('⍎'=1↑⍵)↓⍵}⎕IO⊃⎕DM
          :EndTrap
      :Else
          :Trap DEBUG↓mon←0
⍝ We need to know the calling space. If we were called by <UCMD> this is in 'THIS'
              :If t←'⎕SE.UCMD'≡2⊃⎕XSI,0 ⍝ called by ⎕SE.UCMD?
              :AndIf 9=⎕NC'⎕SE.THIS'
                  cs←⎕SE.THIS ⍝ use what we know
              :Else
                  cs←RSI 1+t                          ⍝ could be a user call
              :EndIf
              'c'⎕NS'' ⍝ we need a NAMED ns to get a full pathname
⍝ If we are debugging we use the object in the ws instead if possible
              :If DEBUG>0
              :AndIf 9=cs.⎕NC t←(com←ObjName⊃,list[i;]),'.SALT_Data'
              :AndIf list[i;FullName]≡⍬⍴splitName(cs⍎t).SourceFile
                  spc←cs⍎com ⋄ ⎕←1⌽'"* Debugging ws object "',com
                  spc.##.THIS←cs ⍝ THIS must be defined for some cmds
              :Else
                  spc←⎕SE.SALT.Load'"',(FullName⊃,list[i;]),SALTEXT,'" -target=',⍕c
              :EndIf
              cmd←,spc.List ⋄ i←(,¨cmd.Name)⍳⊂com←Name⊃,list[i;]
              :If ''≢rules←⌽splitRule cmd[i].Parse
⍝ We have to prepare the parser for the number of arguments
                  (2⊃rules){(0<⍴⍺)/⍵,⍺}←'nargs='
                  arg←(⎕NEW ⎕SE.Parser rules).Parse arg
              :EndIf
     
              :If mon←(9.1=⎕NC⊂'spc')∧(MONITOR>0)∧0<⍴names←spc.⎕NL-4.1 3.1
                  (¯1+⍳999)∘spc.⎕MONITOR¨names ⋄ mon←¯1
              :EndIf
              c.THIS←cs   ⍝ ensure we have a reference to the calling environment
              c.RIU←⍬⍴rgn ⍝ and let know if the result is used
              c.SourceFile←FullName⊃,list[i;]
              c.DMX←DMX
     go:      R←spc.Run com arg
              :If 0≠mon
                  i←(0<+/∘,¨0 1∘↓¨i)/(spc.⎕CR¨names){⍺ ⍵}¨i←spc.⎕MONITOR¨names
                  ⍎name,'←i',(2=⎕NC name)/' MergeMD ',name←MONITORNAME ⍝ set or add data
              :EndIf
          :Else
              R←⎕TC[1],'* Command Execution Failed: ',mon{⍺=1:'Cannot monitor classes' ⋄ ('⍎'=1↑⍵)↓⍵}⎕IO⊃⎕DM
              MONITOR∧←mon≠1 ⍝ turn it off to prevent further problems
          :EndTrap
      :EndIf
    ∇

      MergeMD←{0∊⍴ra←⍵:⍺ ⋄ ∧/keep←~merge←⊃∊/(new old)←1⊃¨¨⍺ ⍵:⍵,⍺ ⍝ all new
          ra[2,⍨¨old⍳merge/new]+←0,¨0 1∘↓¨2⊃¨merge/⍺ ⋄ ra,keep/⍺
      }

    ∇ msg←BootSpice;ev;t;v1112
    ⍝ Used to set up Spice cmd line which is obsolete since V12.1.
      '⎕se.cbbot.bandsalt'⎕NS''
     
    ⍝ We now add the necessary elements to make this work (Parser and Utils should be there already)
      ⎕SE.cbbot.bandsalt.Spice←⎕SE.⎕NS''
      msg←0 0⍴t←⍕1↑⍴2⊃(CMDDIR ⎕SE.cbbot.bandsalt.Spice.List)←1 GetSpiceList'*'
      Spice'usetup init'
    ∇

    ∇ (cmddir list)←{tell}GetSpiceList names;z;t;ns;nc;folder;show1
⍝ Retrieve the list of all Spice commands
      :If show1←326∊⎕DR names
          folder←1↑(cmddir names)←names
      :Else
          folder←ClassFolder∘''¨(cmddir←⎕SE.SALT.Settings'cmddir')splitOn';'
      :EndIf
      z←↑⍪/⎕SE.SALT.List¨'"',¨folder,¨⊂FS,names,'" -rec -raw -full=2'
      z←(0=,⊃⍴¨z[;1])/z[;2]
      :If ~show1
          z∪←(BootPath'Spice',FS)∘,¨'Spice' 'SaltInSpice' 'NewCmd' 'rebuild'  ⍝ always there
      :EndIf
      list←⍬
      :If 0=⎕NC'tell'
          tell←0
      :EndIf
⍝ Spice keeps track of the commands in the Spice folder
      :For t :In z
          :Trap DEBUG↓0
              ns←⎕SE.SALT.Load'"',t,'.dyalog" -noname'
              :If 9=⎕NC'ns'
              :AndIf 3=⌊|ns.⎕NC⊂'List'
                  nc←{⍵.(Name Desc Parse Group)←,¨⍵.(Name Desc Parse Group) ⋄ ⍵}¨,ns.List
                  nc.ObjName←⊂{⍵↑⍨-⊥⍨'.'≠⍵}⍕ns
                  nc.FullName←⊂t
                  list,←nc
              :EndIf
          :Else
              tell{⍺:⎕←⍵}'Error loading Spice Commands from ',t,': ',⎕IO⊃⎕DM
          :EndTrap
      :EndFor
      →(⍴list)↓0
⍝ Remove duplicates, if any
      list←((⍳⍴t)=t⍳t←lCase¨list.Name)/list
      list←⊃list.(Name Group Desc Parse ObjName FullName) ⍝ get rid of namespaces
    ∇

⍝           THE UCMD fn

    ∇ r←{THIS}UCMD Input;rlb;in;nma;ur;dcl
⍝ User Command Processor. The result may be assigned to a global variable.
⍝ Most of the implementation is in ⎕SE.SALTUtils.Spice; this function only implements
⍝ "assignment syntax", i.e. ]varname←ucmd   and   ]⎕←ucmd
     
      rlb←{(+/∧\' '=⍵)↓⍵} ⋄ →(⍴in←rlb Input)↓0
     
      :If 0=⎕NC'THIS' ⋄ THIS←⍎⎕IO⊃⎕NSI ⋄ :EndIf    ⍝ record calling environment
     
⍝ ]var←Cmd accepted; see if it makes sense if present. No spaces allowed.
      nma←(~' '∊nma)/nma←(⌽∨\⌽<\('←'=in)>∨\in∊'''"')/in
      'assignment syntax error'⎕SIGNAL 2⍴⍨('⎕←'≢nma)∧(1<⍴nma)>0 2 9∊⍨THIS.⎕NC ¯1↓nma
      →(⍴in←(⍴nma)↓in)↓0                           ⍝ ]v← is the same as ]
      dcl←'*'∊4⍴⎕STACK                             ⍝ were we at Desk Calculator Level?
     
      :If (0 2∊⍨10|⎕DR ur)∧1∊⍴⍴ur←(dcl≤⍴nma)⎕SE.SALTUtils.Spice in
      :AndIf (2↑ur)≡⎕AV[1+⎕IO],'*'                 ⍝ error signature
          ur ⎕SIGNAL dcl↓911 ⋄ ⎕←ur ⋄ →0           ⍝ signal error only if called by a program
      :EndIf
     
      →(1=⍴nma)⍴0                                  ⍝ ]←cmd    discard result
      :If 0∊⍴nma ⋄ r←ur                            ⍝ no '←'   return result by this fn
      :ElseIf '⎕←'≡nma~' ' ⋄ {⎕←⌽rlb⌽⍵}¨↓⎕FMT ur   ⍝ ]⎕←      display line by line
      :Else ⋄ {THIS⍎nma,'⍵'}ur                     ⍝ ]xx←cmd  store result in xx in calling namespace
      :EndIf
    ∇

⍝ =====================   Boot   fns   =========================

    BootPath←{({⍵↓⍨-'/\'∊⍨¯1↑⍵}getEnvir'SALT'),FS,⍵}

    ∇ BootSALT;bl
⍝ Bring in all necessary code to run SALT
      '⎕se.Dyalog'⎕NS''               ⍝ make sure this one is there
      bl←BootPath''
      bl BootLib'Dyalog.Utils'
      bl BootLib'Parser'
      bl BootLib'UnicodeFile'         ⍝ unused by SALT itself but used by other code
      bl BootLib'SALTUtils'
      bl BootLib'SALT'
      ⎕SE.SALTUtils.EditorFix'Start'  ⍝ set callback in the NEWly defined ns
      ⎕SE.⎕FX ⎕CR'UCMD'               ⍝ define <UCMD> fn
     
      :If ∨/'1yY'∊getEnvir'SALT\AddSpice'
          BootSpice
      :EndIf
     
⍝ Check for autostart (dyapp= on command line)
      :If 0≠⍴bl←getEnvir'dyapp'
          ⎕←'Booting ',bl
          ⎕SE.SALT.Boot{Q,⍵,Q←1↑'"'''~⍵}bl
      :EndIf
    ∇

    splitOnNL←{(13 GetUnicodeFile ⍵) splitOn ⎕ucs 10}

    ∇ loc BootLib target;tgt;file;t;name
⍝ Bring in a single Unicode file
      file←loc,'core',(¯1↑loc),target{⍵,⍨(-⊥⍨'.'≠⍺)↑⍺}'.dyalog'
      t←splitOnNL file
⍝ The name of the object produced MUST match the target name:
      (tgt name)←'.'splitLast target
      tgt←⎕SE⍎tgt,(0∊⍴tgt)/'##'
      {}÷target≡4↓⍕name←tgt.⎕FIX t
      name.SALT_Data←⎕SE.⎕NS''
      name.SALT_Data.(SourceFile LastWriteTime Version)←file'unknown' 0
    ∇

⍝ =====================   EDITOR fns et al  =========================

    ConfirmEdit←1 ⍝ default, reset by <Settings>

    DF←{(⍵.⎕DF df){⍵}⍕⍵{⍺}df←⍵.⎕DF ⎕NULL}  ⍝ cover non-standard ⎕DFs

    ∇ EditorFix args;file;r;t;ns;target;nc;i;nss;ss;id;⍙;sd;name;fixed;df
      :Access Shared Private
⍝ Callback fn for the editor
     
      →0/⍨∨/'0nN'∊getEnvir'SALT\AddSALT' ⍝ has SALT been disabled?
      :If 1≡≡args
          args←0 args   ⍝ pad to get min 2
      :EndIf
     
      :Select t←2⊃args
     
      :Case 'Fix' ⍝ editor callback: [3] source, [4] ns, [5] name
⍝ When the editor exits we grab its SALT data for the AfterFix event
⍝ and we store it in ⎕se.Dyalog along with source ID
     
⍝ We can edit something IN a class. If that object is not itself SALTed
⍝ we update the entire class
          nss←⍴¨ss←⊂'⍝SALTSource="'
          :If ∨/3 4∊nc←⌊|⎕NC⊂id←(DF target←args[4]),'.',5⊃args
          :AndIf ~0∊⍴t←fnData args[3 5]             ⍝ SALTed fn?
              ⎕SE.Dyalog.⍙←id t 0
          :ElseIf 2∊nc
          :AndIf ~0∊⍴t←varData⊂id                   ⍝ SALTed var?
              ⎕SE.Dyalog.⍙←id t 0
          :Else   ⍝ did we edit something IN a SALTed class?
              :If (9=target.⎕NC'SALT_Data')∨ss∊nss↑¨{0::⍬ ⋄ ⎕SRC ⍵}target
                  args[5]←⊂{(1-(⌽⍵)⍳'.')↑⍵}DF target⍝ So "move up"
                  args[4]←args[4].##
                  target←4⊃args
              :EndIf
     
              :If (nc←⌊|target.⎕NC⊂name←5⊃args)∊3 4
                  :If ~0∊⍴t←fnData args[3 5]        ⍝ SALTed fn?
                      ⎕SE.Dyalog.⍙←id t 0
                  :EndIf
              :ElseIf 2∊nc
                  :If ~0∊⍴t←varData⊂id              ⍝ SALTed var?
                      ⎕SE.Dyalog.⍙←id t 0
                  :EndIf
              :Else
                  :If 9=target.⎕NC ns←name,'.SALT_Data' ⍝ NS which has a SALT_Data
                      t←target.⎕OR ns                   ⍝ grab a copy
                      :If df←9=nc
                          df←⍕target⍎name               ⍝ and its display form
                      :EndIf
                      ⎕SE.Dyalog.⍙←id(⎕NS t)df          ⍝ being careful not to keep the original
                  :ElseIf (⍴t)≥i←(nss↑¨t←{0::⍬ ⋄ ⎕SRC ⍵[4]⍎5⊃⍵}args)⍳ss
                      ⍙←⎕NS''
                      ⍙.SourceFile←{(¯1+⍵⍳'"')↑⍵}{(⍵⍳'"')↓⍵}i⊃t
                      ⍙.Version←0
                      ⍙.LastWriteTime←⍕lastWrTime ⍙.SourceFile
                      ⍙.GlobalName←''
                      ⎕SE.Dyalog.⍙←id ⍙ 0
                  :EndIf
              :EndIf
          :EndIf
     
      :Case 'AfterFix'
⍝ After an object has been fixed successfully we add the SALT details if any
⍝ which we stored in ⎕se.Dyalog
          target←fixed←df←0
          :If 2=⎕NC sd←'⎕se.Dyalog.⍙' ⍝ they were kept in this global in this ns
          :AndIf (,3)≡⍴t←⍎sd          ⍝ we need a specific signature
          :AndIf ((DF target←4⊃args),'.',5⊃args)≡1⊃(id ⍙ df)←t
⍝ If ⍙ contains 'Name', it represents the name of the fn fixed if it was the case.
⍝ If there is no such name then a class was fixed.
              :If 0=⍙.⎕NC'Name' ⍝ is it a class in a scripted ns? won't work properly in V11 or 12
⍝ ns in non scripted nss are OK
                  :If 9=target.⎕NC 5⊃args
                      :If {16::0 ⋄ 1{⍺}⎕SRC ⍵}target
                      :AndIf ≤Ver 12
                          ⎕←'*** Cannot fix a sourced ns within another sourced ns in V<12.1 ***'
                          ⎕SIGNAL 11 ⍝ to prevent editor from closing
                      :EndIf
                      target←target⍎5⊃args
                  :EndIf
                  target.SALT_Data←target.⎕NS ⎕OR'⍙' ⍝ needed in 12.1
              :EndIf
              fixed←Fixed target ⍙
              {}target.⎕DF⍣(0≢df)+df      ⍝ restore ⎕DF if was present
          :EndIf
          ⎕EX sd ⍝ prevent accidents
          →L10/⍨fixed∧'⎕SE.SALTUtils'≡⍕target
     
      :CaseList 'Start' 'chk'
     L10: SE_Editor←'⎕SE'⎕WG'Editor'
          SE_Editor.⎕WX←1
          :If t≡'chk'
              ⎕←'SE_Editor'⎕WG'Event'
          :EndIf ⍝ sometimes it goes away
          'SE_Editor'⎕WS'Event'('Fix' 'AfterFix')(⎕IO⊃⎕XSI)
     
      :Case 'End'
          'SE_Editor'⎕WS'Event'('Fix' 'AfterFix')0
          ⎕EX'SE_Editor'
     
      :EndSelect
    ∇

    ∇ done←Fixed(ns ⍙);name;source;z;t;r;new;upd;version;sourcefile;n;here;∆;isns;path;ext;maxv;confirm;last;prev;buttons;ync;isfn;s0;nofile;mode
      :Access Shared Private
    ⍝ Called after object has been fixed to update the source file
      nofile←1 ⋄ isfn←done←0 ⋄ z←,⊂''
      :If isns←0∊⍙.⎕NC'Name' ⍝ Class? Namespaces have no 'Name' var
          source←⎕SRC ns ⋄ name←DF ns
      :ElseIf isfn←4 3∨.=ns.⎕NC name←⍙.Name
          source←remTag ns.⎕NR name  ⍝ remove tag line for fns
      :Else ⍝ it's a variable, hopefully palatable
          source←name ∆VCR⍎⍙.Pathname
      :EndIf
      :Trap 22 ⍝ the file may be gone
          z←fixTabs splitOnNL ⍙.SourceFile ⋄ nofile←0
      :EndTrap
      :If ('<APLScript>'startsWith⍨1⊃z)∧':Class PageClass'startsWith⍨1⊃source
          source←z MakeAPLScript source
      :EndIf
     
      :If source≢z ⍝ Any changes?
      :AndIf (rlb¨source)≢rlb¨z ⍝ we shouldn't have to do this
     
⍝ This is an update; there are 2 cases:
     
⍝ 1. we are adding a new version: we use the next available number
⍝    We have to be careful not to add a new version from an outdated version (e.g. we loaded an
⍝    old ws containing an outdated version). For this we check that the version # is exactly the
⍝    same as the highest verno. In the case where we actually requested to reload an old version
⍝    it will be negative to denote that fact (i.e. Load will negate the verno)
⍝ In the case where a brand new version is added (we edited the very LAST version) the prompt includes
⍝ an option to overwrite the last version.
     
⍝ 2. we are replacing an existing file: we confirm if confirmation not disabled
⍝    If we attempt to overwrite with an old version we will detect it by comparing the timestamps
     
          z←s0←sourcefile←⍙.SourceFile
          :If new←last←0≠version←maxv←⍙.Version ⍝ case #1
            ⍝ There should be a version # but just in case...
              (t n ext)←splitName 2⊃(path t)←FS splitLast sourcefile
              maxv←⌈/0,path ListVersions t,'.',ext ⍝ find last verno
              last←version≥maxv ⍝ if it is > it must be because we used RemoveVersions
              version←1+maxv
              (s0 sourcefile prev)←(path,FS,t)∘,¨((⊂''),'.',¨⍕¨1 0+maxv),¨⊂'.',ext
          :EndIf
     
         ⍝ We can automate confirmation by setting 'ConfirmEdit' to be 1 (YES)
         ⍝ The 2nd element of 'ConfirmEdit', is present, is the answer to the following
         ⍝ 3 questions.
          :If ~1↑(confirm ync)←2↑ConfirmEdit  ⍝ has confirmation been disabled?
              ync←3⍴ync ⍝ used for answering the next 3 questions automatically
            ⍝ Confirmation required, prepare the question
              upd←⊂('Update' 'Create a new'⊃⍨1+new),' source file for "',name,'"?'
              t←(nofile<new>last)/'NOTE: *** this is NOT the latest version!'
              t←t,nofile/'NOTE: *** the original file is no longer found!'
              mode←'overwritten.' 'created.'⊃⍨1+new∨nofile
              t←t('If you choose YES, file <',sourcefile,'> will be ',mode)''
              buttons←'Yes' 'No',last/⊂'Cancel'
              :If last←last>nofile
                  t,←⊂'If you choose NO, file <',prev,'> will be rewritten'
                  t,←'' 'If you choose CANCEL, the changes won''t be filed'
              :Else
                  t,←⊂'If you choose NO, the changes won''t be filed'
              :EndIf
          :OrIf (2+last)>ync←msgBox('You have modified ',⍕name)(upd,t)'Query'buttons
     
              :If confirm<×DEBUG
                  ⎕←'* auto response to "You have modified...":',1↑ync
              :EndIf
              :If last∧2∊1↑ync
                  sourcefile←prev ⋄ version←maxv
              :EndIf
     
            ⍝ Make sure the file is not stale.
            ⍝ NOTE: we consider the file up to date if the version is > the one on disk
            ⍝ which may happen after running REMOVEVERSIONS
              :If (maxv>n)∧0<n←⍙.Version
                  :If confirm
                      t←('This is version ',(⍕n),', the latest is ',⍕maxv)('Do you still want to save it as V',(⍕maxv+1),'?')
                      →0⍴⍨1≠msgBox'*** THIS IS NOT THE MOST RECENT VERSION !!!'t'Warn'
                  :Else
                      :If ×DEBUG
                          ⎕←'* auto response to "This is version X":',t←1↑1↓ync ⋄ →0/⍨t≠1
                      :EndIf
                  :EndIf
            ⍝ File is not new: ensure it hasn't been updated outside
              :ElseIf ~nofile
              :AndIf ⍙.LastWriteTime{⍺≢⍵:6<⍴⍺∪⎕D ⋄ 0}t←⍕lastWrTime ⍙.SourceFile   ⍝ not same as we loaded?
                ⍝ If no timestamp is found the file is probably gone, all we need to do is rewrite (with permission)
                ⍝ If its folder is also gone we can't do it and we simply tell the user when we fail to write the file.
                  :If confirm
                      t←{0∊⍴⍵:'No timestamp found for original file!' ⋄ 'Now dated ',⍵}t
                      t←⍙.SourceFile''('Was dated ',⍙.LastWriteTime,' when loaded...')(t)'Proceed anyway?'
                      →0⍴⍨1≠msgBox'Source file timestamp has changed...'t
                  :Else
                      :If ×DEBUG
                          ⎕←'* "Was dated" auto response:',t←¯1↑ync
                      :EndIf
                      →0/⍨t≠1
                  :EndIf
              :EndIf
              z←mergeTxt('⍝:Pragma Line 1'startsWith⍨1⊃source)↓source
              :Trap 22
                  z PutUTF8File t←sourcefile
                  :If new
                      z PutUTF8File t←s0
                  :EndIf
              :Else
                  ⎕←'*** Unable to save <',t,'>, SALT tag is removed'
                  ⎕←'*** Use ]SAVE to save the object manually'
                  :If isfn
                      ns.⎕FX source
                  :ElseIf isns
                      ns.⎕EX'SALT_Data'
                  :Else
                      1 varData⊂(⍕ns),'.',name
                  :EndIf
                  →0
              :EndTrap
              done←1
              :If isns
                  ⍙←ns.SALT_Data ⍝ shorthand
                  ⍙.LastWriteTime←⍕lastWrTime ⍙.SourceFile←sourcefile ⍝ update Timestamp
                  ⍙.Version←version
                  :If DEBUG>1
                      ⎕←'* New Tag info: ',⍙.(LastWriteTime SourceFile Version)
                  :EndIf
              :Else
                  SetDelta(ns name,isfn↓⊂source)sourcefile version ⍬ 0
              :EndIf
          :EndIf
      :EndIf
    ∇

    mergeTxt←{(-⍴L)↓⊃,/(⊂⍣(1≡≡⍵)+⍵),¨⊂⎕ucs L←LINDEL}

    ∇ r←file MakeAPLScript class;html;i;script;m;eAS
⍝ Merge PageClass source with apls file contents
      html←(i←((⍴eAS)↑¨file)⍳eAS←⊂'</APLScript>')↓file ⍝ Extract the non-APLScript part
      script←1↓(i-1)↑file
      script←({'<'∊1↑⍵~' '}¨script)/script ⍝ statements beginning with <
      :If (⍴class)≥i←class⍳⊂'⍝--- APLS Code Begins ---'
          class←¯2↓i↓class
          m←class∧.=¨' '
          class←(+/∧\m)↓(-+/∧\⌽m)↓class
      :EndIf
      r←(eAS~¨'/'),script,class,eAS,html
    ∇

⍝ ===== SALT utilities

    ∇ {rc}←Forget(folder name version keeplast newlast noprompt);prev;t;f;list;dotVer;next;ok;S;nf;ext;stem;last;Ver
⍝ Forget Backup Versions
      :Access Shared Private
      ok←1 ⋄ dotVer←{(0<⍵)/'.',⍕⍵} ⋄ S←{(1=⍴⍺)↓'s',⍵} ⍝ add 's' to plural
      :If 12<nf←⍴t←list~keeplast/⌈/list←{⍵[⍋⍵]},version
          t←t[1],(⊂'...'),¯1↑t
      :EndIf
      (stem f ext)←splitName name ⋄ Ver←'Version',t S' ',⍕fmtVersion¨t ⋄ ext←'.',ext
      :If 0<nf
          t←('Forget Versions of ',name)(('Confirm deletion of ',(⍕nf),' version',list S':')''Ver)'Warn'
      :AndIf ok←t{⍵:1 ⋄ 1=msgBox ⍺}noprompt
          :For t :In list
              fileErase f←(folder,FS,stem,dotVer t),ext
          :EndFor
⍝ This should go to the stderr stream
          ⎕←(⍕nf),' version',list S' deleted.'
      :EndIf
      →ok↓0×rc←nf⌈⍣ok-1 ⍝ return -1 if user aborted
⍝ If a new last version results of the deletions we must make a copy to the non-versioned file
      :If newlast∨.<version ⍝ then the old current version has been deleted
      :AndIf keeplast∨newlast>0
          f←folder,FS,stem,ext ⋄ last←(folder,FS,stem,dotVer newlast+keeplast),ext
          fileCopy/keeplast⌽last f
      :EndIf
    ∇

    ∇ SetDelta(ref name version globalname xml);prev;nr;targ;ll;t;d;src;lw
      :Access Shared Private
      lw←⍕lastWrTime name
      :If 3∊⍴ref ⍝ variable
          (targ ref src)←ref
          varData((DF targ),'.',ref)name version lw(calcCRC src)xml
      :ElseIf 2∊⍴ref ⍝ function
          (targ ref)←ref ⋄ ll←⍬
          nr←,condEncl remTag targ.⎕NR ref
          :If ∨/t←0.3 0.2=1|targ.⎕NC⊂ref
⍝ One liners must be split to insert the tag line and idioms must be assigned
              (1⊃nr),⍨←t[1]/ref,'←' ⍝ patch idioms
              :If ∧/((⍴nr)⊃nr)∊' }'
                  ll←¯1↑nr ⋄ nr←¯1↓nr
              :Else
                  nr[⍴nr]↓⍨←¯1 ⋄ ll←⊂' }'
              :EndIf
          :EndIf
          d←'§' ⋄ nr,←⊂'⍝∇⍣',d,name,d,(⍕version),d,lw,d,calcCRC nr,ll
          targ.⎕FX nr,ll ⍝ refix fn with info
      :Else ⍝ must be a ref
          prev←¯1
          :If 2∊ref.⎕NC'SALT_Data.Version'
              prev←ref.SALT_Data.Version
          :EndIf
          ref.SALT_Data←ref.⎕NS''
          ref.SALT_Data.PreviousVersion←prev
          ref.SALT_Data.SourceFile←name
          ref.SALT_Data.Version←version
          ref.SALT_Data.LastWriteTime←lw
          ref.SALT_Data.GlobalName←globalname
          ref.SALT_Data.CRC←calcCRC ⎕SRC ref
      :EndIf
    ∇

    ∇ dstname CopyNs source;scrref;dref;vars;nss;ns;cls;val;name;b;src;fn;s
⍝ Copy contents of a NAMESPACE into a new one
⍝ This fn works 2 ways:
⍝ 1: with a source reference which is to be recreated wo source to the target
⍝ 2: with a source NAMED which is NOT to be recreated if its display form does not match
      :If 2=⍴,scrref←source
          name←⍕scrref←⍎/source
          →0 if name≢(⍕1⊃source),'.',2⊃source ⍝ skip refs
      :EndIf
      ⎕EX dstname         ⍝ ensure not there
      dref←⍎dstname ⎕NS'' ⍝ create new namespace & get its reference
      dref.(⎕IO ⎕ML ⎕WX)←scrref.(⎕IO ⎕ML ⎕WX) ⍝ ⎕vars
      vars←scrref{0∊⍴⍵:0 2⍴0 ⋄ ⍵,[1.5]⍺.⍎¨⍵}scrref.⎕NL-2
      :For name val :In ↓vars
          ⍎dstname,'.',name,'←val'
      :EndFor
      :If ∨/b←1 3∊⍨10|⎕DR¨dref.⎕FX¨scrref.(⎕NR¨⎕NL-3 4)
⍝ OK, some fn didn't fix, probably because it is a ref of some sort
⍝ If it happens to be made with ∘ there is nothing we can do, OTOH
⍝ we may be able to rebuild it like this:
          'Invalid fn'⎕SIGNAL 11/⍨∨/'∘'∊¨src←scrref.⎕NR¨name←b/scrref.⎕NL-3 4
          :For fn s :InEach name src
              dref⍎fn,'←',⍕s ⍝ we can't do them all at once
          :EndFor
      :EndIf
      :For cls :In scrref.⎕NL-vars←9.5 ⍝ Interfaces
          dref.⎕FIX scrref.{⎕SRC⍎⍵}cls
      :EndFor
⍝ Naive redefinition: if base classes are present this will fail
      :For cls :In scrref.⎕NL-9.4
          dref.⎕FIX scrref.{⎕SRC⍎⍵}cls
      :EndFor
      :For ns :In (scrref.⎕NL-9.1)~⊂'SALT_Data' ⍝ Namespaces
          (dstname,'.',ns)CopyNs(scrref ns)
      :EndFor
     ⍝
    ∇

    ∇ cleanWS tgt;tgt;lst;sons
⍝ This fn is used to remove all traces of linking startin at tgt
      →0↓⍨{0::1 ⋄ {0}⎕SRC ⍵}tgt ⍝ skip scripted nss
      :If 0<⍴lst←tgt.⎕NL-3 4
          {}tgt remTag{⍺.⎕FX ⍺⍺ ⍺.⎕NR ⍵}¨lst
      :EndIf
      :If 0<⍴lst←tgt.⎕NL-9
          sons←tgt~⍨tgt⍎¨lst ⋄ sons.⎕EX⊂'SALT_Data'
⍝ Recurse thru non-scripted children
          :If 0<⍴sons
              cleanWS¨sons
          :EndIf
      :EndIf ⍝ all nss
      #.⎕EX'SALT_Var_Data'
    ∇

    ∇ r←folder ListVersions name
      :Access Shared Private
      r←⎕IO⊃folder ListVersionsTS name
    ⍝ Version 0 equates the highest Vno if present
      r←r~(0<⌈/r)/0 ⍝ remove it if versioning in effect
    ∇

    ∇ (r ts)←folder ListVersionsTS name;files;se;ext;u
      :Access Shared Private
⍝ Return version numbers and the timestamp of the associated files.
⍝ 'name' is a name always followed by an extension ONLY (no verno)
⍝ 0 as verno represents a file with no version #.
      (name r ext)←splitName lCase⍣WIN⌷name ⍝ lowercase for Windows
      se←-⍴ext←'.',ext,(0∊⍴ext)/1↓SALTEXT
      folder←folder,(FS=¯1↑FS,folder)↓FS
      r←⍬
      :If 0<⍴⎕IO⊃(ts files)←2 4⊃¨⊂'a'Dir folder,name,'*',ext
⍝ There is a possibility that Dir reported too many names:
      :AndIf 0<⍴⎕IO⊃(files ts)←((name,'.')∘≡¨lCase⍣WIN¨(1+⍴name)↑¨files)∘/¨files ts
          u←(r⍳r)=⍳⍴r←{(('.'∊⍵)>' '∊v)×+/2⊃⎕VFI v←2⊃'.'splitLast ⍵}¨se↓¨files
          (r ts)←u∘/¨r ts
      :EndIf
    ∇

    ∇ path←specialName name;rp;nwb;pfs
⍝ Change any [name] into path
      :Access Shared Private
      →0↓⍨'['=1↑path←name
      path←getEnvir uCase nwb←((rp←name⍳']')↑name)~'[]' ⍝ uCase for Unix
      ('Folder cannot be resolved: ',name)⎕SIGNAL 922⍴⍨0∊⍴path
      rp+←FS∧.=(pfs←¯1↑path),1↑rp↓name
      :If 'dyalog'≡lCase nwb ⍝ 'Dyalog' needs the 'bin' folder removed
          path↓⍨←¯4×(FS,'bin')≡¯4↑(-pfs=FS)↓path
      :EndIf
      path,←rp↓name
    ∇

    ∇ folder←root ClassFolder folder;special;and;rooted;rs;stem;nocl;rp
⍝ Produce full path by merging root and folder name
      :Access Shared Private
⍝ [xxx] in folder names resolved using GetEnvironment including [Dyalog]
      folder↓⍨←-FS=¯1↑folder                  ⍝ remove last \
      →((rp≢FS,'*')∧∨/(FS,':')=rp←2↑folder)⍴0 ⍝ is it already formatted properly?
      and←{((-FS=¯1↑⍺)↓⍺),(0<⍴⍵)/FS,('/\'∊⍨1↑⍵)↓⍵}
      special←''
     
⍝ Special names treated here
      :If '['=1↑folder
          folder←0/root←folder
      :EndIf
      :If '['=1↑root
          special←specialName(rp←root⍳']')↑root
          folder←special and(rp↓root)and folder
      :Else
          folder←root and folder
      :EndIf
    ∇

    ∇ (dir path files)←name locateIn folders;dir;nfs;path;files;t;empty
⍝ Find the 1st occurence of file in list of folders
⍝ Return the folder it was found in the list provided. the full path and the files found
      :Access Shared Private
      empty←' '∧.=name ⍝ this will force return on the first dir in the list
      :For dir :In folders
          :If empty∨0<⍴⎕IO⊃files←'a'Dir path←dir ClassFolder name
              :If ∨/'?*'∊2⊃t←FS splitLast path
                  path←1⊃t
              :EndIf
              →0
          :EndIf
      :EndFor
      dir path files←'' '',⊂,¨4⍴⊂⍬ ⍝ default not found
    ∇

    ∇ nr←remTag nr
⍝ Remove any tag line in the fn
      nr/⍨←'⍝∇⍣'∘≢¨3∘↑∘rlb¨nr
    ∇

    ∇ ns←fnData arg;lns;line;data;⎕ML;ln;b
⍝ Fetch data in fn: sourcefile:version:lastwrtime:crc
⍝ arg is (source name)
      ns←⍴⎕ML←1 ⋄ →0↓⍨ln←⍴lns←1⊃arg
      →0↓⍨1≡≡line←ln⊃lns
      →0↓⍨ln←ln-'}'=¯1↑line
      →0↓⍨∨/b←'⍝∇⍣'∘≡∘(3∘↑)∘rlb¨lns            ⍝ do we have such a line?
      →0↓⍨4=⍴data←1↓¨('§'=line)⊂line←(b⍳1)⊃lns ⍝ is it kosher?
      data[2]←2⊃⎕VFI 2⊃data
      ns←⎕NS'' ⋄ ns.(SourceFile Version LastWriteTime CRC Name)←data,arg[2]
    ∇

    ∇ {ns}←{rem}varData arg;gv;gd;set;gns;b
    ⍝ Fetch or Set data in ws for variable given as arg
      set←1<⍴,arg ⋄ ns←⍬ ⋄ gns←¯3↓gv←'#.SALT_Var_Data.VD'          ⍝ return ⍬ if not found
      :If 0=⎕NC gv                                                 ⍝ initialize if absent
          →set↓0
          ⍎gns,'←#.⎕ns⍬⋄',gv,'←0 6⍴0'
      :EndIf
      gd←⍎gv                                                       ⍝ fetch whole data
      :If 6>¯1↑⍴gd ⋄ gd←gd,0 ⋄ :EndIf                              ⍝ account for pre V1.27 versions
      :If 0=⎕NC'rem'
          rem←0
      :EndIf
      :If ~∨/b←gd[;1]∊⍬⍴arg
          →(rem≥set)/0
          b,←~gd⍪←0
      :EndIf
      ns←⎕NS'' ⋄ ns.Name←{⍵↑⍨-⊥⍨'.'≠⍵}⎕IO⊃,arg
      ns.(Pathname SourceFile Version LastWriteTime CRC Format)←,b⌿gd     ⍝ find actual values
      :If rem
          gd←(~b)⍀gd
      :EndIf
      →set↓0
      gd[b⍳1;]←arg ⋄ ⍎gv,'←gd'                                     ⍝ reset globally
    ∇

    ∇ r←name ∆VCR object;xml ⍝ create a visual char represention of a variable
      xml←0
      :If 2≡|≡name
          (name xml)←name
      :EndIf
      :If xml
          r←'⌷',name,'←',⎕SE.Dyalog.Utils.toXML object
      :Else
          r←'⌷',name,'←',⎕SE.Dyalog.Utils.repObj object
      :EndIf
    ∇

    valWithRef←{1∊{9∊⎕NC'⍵':1 ⋄ ≡⍵}¨∊⍵} ⍝ does this value contain a ref or ⎕OR?

    ∇ crc←calcCRC vtv;⎕ML;n;DH;⎕IO
    ⍝ Compute a CRC from a list of text vectors
      n←⍴DH←⊃,/LU ⋄ ⎕IO←1 ⋄ ⎕ML←1
      crc←DH[1+(6⍴n)⊤(⍳⍴crc)+.×⎕AV⍳crc←∊vtv]
    ∇

    SALTEXT←'.dyalog' ⍝ DOT MUST be part of it

    ∇ (stem ver ext)←{def}splitName name;d;n
⍝ Split filename into constituent parts
      ⍎(0=⎕NC'def')/'def←1' ⍝ do we assume a default?
      ext←def/1↓SALTEXT ⋄ ver←0⍴stem←name
      →0⍴⍨∧/d←(name≠'.')∨d≠¯1↑d←+\name∊'/\' ⍝ any dots?
      stem←name↓⍨n←-1+⍴ext←(-⊥⍨d)↑name ⋄ n←-⊥⍨n↓d
      →0↓⍨n{(⍺>-⍴⍵)∧(' '∊⍵)<(,1)≡⎕IO⊃⎕VFI ⍺↑⍵}stem
      stem←stem↓⍨-1+⍴ver←n↑stem
    ∇

    ∇ name←remExt name;d ⍝ remove whatever extension after the last DOT (.)
      →0↓⍨∨/d←(name='.')∧⌽∧\~⌽name∊'/\'
      name↓⍨←-1+⊥⍨~d
    ∇

⍝ Add official SALT extension if none already there
    addExt←{⍺←SALTEXT ⋄ ∨/(⍵='.')∧⌽∧\~⌽⍵∊'/\':⍵ ⋄ ⍵,'.',('.'=1↑⍺)↓⍺}

    remVerno←{(f v)←'.' splitLast ⍵ ⋄ ∧/(×⍴v),v∊⎕d:f ⋄ ⍵}

    ∇ {(val xml)}←la fixVar src;nam;⍙⍙⍙0;⎕ML;l ⍝ fix the source of a variable in the target space
      nam←(3⊃la)~' ' ⋄ ⎕ML←1
    ⍝ Find the format:
      :If xml←'<'∊1↑(l←2+⍴nam)↓⎕IO⊃,src
          {}⎕FX(⊂'_←⍙⍙⍙0 S'),⊂'_←⎕se.Dyalog.Utils.fromXML ',(⍕l),'↓∊S'
      :Else
    ⍝ Define a fn to recreate the variable
          {}⎕FX(⊂nam,'←⍙⍙⍙0 S;_'),(1<⍴src)↓(1↑⍨⍴src)↓¨src ⍝ remove first line if more than one
      :EndIf
      :Trap 0 ⍝ la is target space, VALUE only, name, protected
          :If ~la[2]
              :If la[4]∧×la[1].⎕NC nam
                  val←'** "',nam,'" is already defined'
              :Else
                  val←nam ⋄ ⍎(DF la[1]),'.',nam,'←⍙⍙⍙0 src'
              :EndIf
          :Else
              val←⍙⍙⍙0 src
          :EndIf
      :Else
          val←'*** Unable to define variable',(1⌽'" "',nam),': ',⎕IO⊃⎕DM
      :EndTrap
    ∇

    ∇ r←fmtVersion v;ab ⍝ arg can be a simple number, or a string
      :Access Shared Private
      ab←{'[',⍵,']'}  ⍝ neg numbers don't get decorators
      r←{3=10|⎕DR ⍵:ab⍣(0<⍵)⍕|⍵ ⋄ (0<⍴,⍵)/ab ⍵}v
    ∇

    ∇ r←fmtDate v
      :Access Shared Private
      r←0 19⍴'' ⋄ →0⍴⍨0∊⍴v
      r←v ⋄ →0⍴⍨isChar v ⍝ empty v is char: skip
      r←'I4,2(</>,ZI2),I3,2(<:>,ZI2)'⎕FMT⊃,6↑¨↓v
    ∇

    ∇ r←lastWrTime w ⍝ Last Write Time for file
      :Access Shared Private
      r←2⊃'a'Dir w
    ∇

    ∇ {r}←CC list;all;b ⍝ Code Coverage. list is the fns to set
     ⍝ list: fns to Code cover or 0 (stop & clear) or -1 (stop)
      :Access public shared
      :Trap 6
          {}CodeCov.Version
      :Else
          ⎕SE.SALT.Load'tools\code\codecov'
      :EndTrap
      all←⎕NL 3 4
      :If 1∊b←¯1 0∊list
          r←CCdata[;2]CodeCov.CC ⎕MONITOR¨CCdata[;1]
          →b[1]/0 ⍝ do not reset numbers if ¯1
          →0,⍬∘⎕MONITOR¨↓all
      :ElseIf list≡1
          list←(~∘' '¨↓all)~'CC' 'UCMD'
      :EndIf
      r←CCdata←CodeCov.CCinit ⎕CR¨⊂⍣(2>≡list)⌷list
      (¯1+⍳999)∘⎕MONITOR¨list ⍝ we can't do this inside
    ∇


⍝ ===== Parsing related fns

    ∇ parms←{patn}∆parse arg;p;⍙;narg
⍝ In this version we make use of the Parser in ⎕SE.
⍝ arg is argument, number of expected args (default any number)
      narg←''
      :If 2∊|≡arg
          (arg narg)←arg
      :EndIf
      narg←'upper prefix=⍙',(0<⍴,narg)/' nargs=',narg
      p←⎕NEW ⎕SE.Parser(patn narg)
      parms←p.Parse arg
      ⍙←1⌽')(',⍕'⍙'parms.⎕NL-2 ⍝ all ⍙ names
      ⍎⍙,'←parms.',⍙ ⍝ assign ⍙ names in current namespace
      parms←parms.Arguments
    ∇

    ∇ r←r ∆default value
⍝ Default CHARACTER STRING r BY value IF IT ≡0
      :Access Shared Private
      →(r≡1)⍴0 ⍝ also a valid value
      →(r≡0)↓mod
      r←value ⋄ →0
      ⍝ r IS not SCALAR 0 - ⍎ IF value IS NUMERIC
     mod:→(' '=1↑0⍴value)/0
      r←2⊃⎕VFI r
    ∇

    ∇ str←∆propagate switches;sw;b;v
⍝ This fn will recreate a string of the switches in order to be passed on another cmd
⍝ e.g. ⎕SE.SALT.Cmd myarg,∆propagate 'VERSION'
      str←''
      :If 1∊b←2=,⊃⎕NC¨sw←'⍙',¨switches splitOn' '
          sw←b/sw
      :AndIf 1∊b←0≢¨v←⍎¨sw
          str←∊(b/v){' -',1↓⍵,(1≢⍺)/'=',⍕⍺}¨b/sw
      :EndIf
    ∇

⍝ ===== Misc Utilities

⍝ String Index with minimal length: 2='aa' 'bb' strIndex ⊂,'b'
    strIndex←{∨/b←⍺∊⍵:b⍳1 ⋄ (1=+/b)×1⍳⍨b←((⍴s)↑¨⍺)∊⊂s←,1⊃,⍵}

    ∇ r←{b}rlb w            ⍝ rem leading blanks
      :Access Shared Private
      :If 0=⎕NC'b'
          b←' '
      :EndIf
      r←(+/∧\w∊b)↓w
    ∇

    ∇ r←{b}rtb w            ⍝ rem trailing blanks
      :Access Shared Private
      :If 0=⎕NC'b'
          b←' '
      :EndIf
      r←(-⊥⍨w∊b)↓w
    ∇

    ∇ r←isChar obj
      :Access Shared Private
      r←0 2∊⍨10|⎕DR,obj
    ∇

    ∇ lu←LU
      :Access Shared Private
      lu←'abcdefghijklmnopqrstuvwxyzàáâãåèéêëòóôõöøùúûäæü' 'ABCDEFGHIJKLMNOPQRSTUVWXYZÀÁÂÃÅÈÉÊËÒÓÔÕÖØÙÚÛÄÆÜ'
    ∇

    ∇ s←lCase s;b;⎕IO;i;n;l;u
      :Access Shared Private
      n←⍴⎕IO⊃(l u)←LU
      →(∨/b←n>i←u⍳s)↓⎕IO←0
      (b/s)←l[b/i]
    ∇

    ∇ s←uCase s;b;⎕IO;i;n;l;u
      :Access Shared Private
      n←⍴⎕IO⊃(l u)←LU
      →(∨/b←n>i←l⍳s)↓⎕IO←0
      (b/s)←u[b/i]
    ∇

    ∇ r←msgBox arg;parms;msgbox;ans;Caption;Text;Mode;Btns;choices;t;hint;evt;Warn;N;No
    ⍝ Return 1, 2 or 3 for buttons pressed
      :Access Shared Private
      (Caption Text Mode Btns)←arg,(⍴arg)↓0 0 'Warn',('Yes' 'No')(⊂'OK')[Warn←1+arg[3⌊⍴arg]∊'Info' 'Msg']
      :If WWIN
          'msgbox'⎕WC'msgBox'Caption Text Mode Btns('Event'(evt←61 62 63)1)
          r←4|evt⍳2⊃⎕DQ'msgbox'
      :Else
          (N No)←Warn⊃¨('nN' 'oO')('N (No)' 'O (OK)')
          choices←'yY',N,(t←3∊⍴Btns)/'cC' ⋄ hint←' Y/',1↓N,(t/'/C'),'? '
          ⎕←((⍴Text),⍳1≠≡,Text)⍴Text
          :While ~1∊ans←choices={⍞←⍵ ⋄ 1↑(⍴,⍵)↓⍞}⍕hint
              ⎕←'** Please answer by Y (Yes) or ',No,t/' or C (Cancel)'
          :EndWhile
          r←(∨/3 2⍴ans)⍳1
      :EndIf
    ∇

    ∇ last←str afterLast chars
      :Access Shared Private
      last←(-⊥⍨~str∊chars)↑str ⍝ string after last char(s)
    ∇

    ∇ r←str splitOn1st chars;p;⎕IO
⍝ Split on 1st occurrence of any chars in str
      :Access Shared Private
      ⎕IO←0 ⋄ p←⌊/(,str)⍳chars ⋄ r←(p↑str)((1+p)↓str)
    ∇

    ∇ yes←string startsWith subs
      :Access Shared Private
      yes←subs≡(⍴subs)↑string
    ∇

    ∇ strs←str splitOn char;⎕ML
      :Access Shared Private
      ⎕ML←1 ⍝ char may be multiple
      strs←1↓¨(strs∊char)⊂strs←(1↑char),str
    ∇

    ∇ r←a splitLast w;p ⍝ Split on last occurrence of a in w
      :Access Shared Private
⍝ Assume in first pos if not there
      p←-(⌽,w)⍳a ⋄ r←(p↓w)((1+p)↑w)
    ∇

    ∇ name←name trimAt char;⎕IO
⍝ Truncates a character vector at the char delimiting byte.
      :Access Shared Private
      ⎕IO←0 ⋄ name↑⍨←name⍳char
    ∇

    ∇ r←isHelp string
      :Access Shared Private
      r←(string~' ')≡,'?'
    ∇

    ∇ r←num r        ⍝ only use ⎕VFI if character
      :Access Shared Private
      :If isChar r
          ((r='-')/r)←'¯'
          r←2⊃⎕VFI r
      :EndIf
    ∇

    ∇ r←condEncl arg ⍝ conditional enclose
      :Access Shared Private
      r←⊂⍣(326≠⎕DR,arg)⌷arg
    ∇

    ∇ _nam Default _val
      :Access Shared Private
      ⍎(0=⎕NC _nam)/_nam,'←_val'
    ∇

    ∇ {ok}←val setPfkey n;s
      :Access Shared Private
      :If ok←0<⍴s←⍕n
      :AndIf ok←0 36≠.≤s←1↑num s
          {}val ⎕PFKEY s
      :EndIf
    ∇

    ∇ r←fixTabs source;⎕IO   ⍝ Change Tabs into N spaces
      ⎕IO←0 ⋄ r←⎕AV[9]{∧/b←⍺≠⍵:⍵ ⋄ b←⍺=l←4 1[b]/⍵ ⋄ (b/l)←' ' ⋄ l}¨source
    ∇

⍝ "BlackJack" Partitionning: cut so +/¨⍴⍵ords is max ⍺ in each partition - works also with lines of text
      BJp←{s←1≡≡⍵ ⋄ l←s{⍺:¯2-/0,b/⍳⍴b←1,⍨' '=⍵ ⋄ 2+∊⍴¨⍵}⍵
          cut←(a↑1)⌹((2⍴a)⍴(1+a←⍴l)↑1)-<⍀b∘.>⍺+¯1↓0,b←+\l
          cut←{¯1↓∊l↑¨⍵}⍣s+cut
          ⎕ML←1 ⋄ cut⊂⍵}

    ∇ r←{options}WidthFit text;sh;blk;sp;pw;⎕ML;⎕IO;n    ⍝ fit a list ⎕PW wide
      :Access Shared Private
      ⍎(0=⎕NC'options')/'options←⍬' ⋄ (pw blk sp)←options,(⍴,options)↓⎕PW,4 1 ⋄ pw←pw-sp
      ⎕ML←⎕IO←1 ⋄ sh←⊃,/⍴¨,¨text
      sh←⊃,/⍴¨text←(blk×⌈blk÷⍨sh+sp)↑¨text ⍝ adjust each word's size
      r←0⍴⊂''
      :While 0<⍴text
          r←r,,/(n←+/pw≥+\sh)↑text ⋄ (text sh)←n↓¨text sh
      :EndWhile
      r←↑r
    ∇

⍝ =====  Directory functions

    ∇ r←isDir w
      :Access Shared Private
      r←1∊⎕IO⊃'a'Dir w
    ∇

    ∇ r←isFile w
      :Access Shared Private
      r←0∊⎕IO⊃'a'Dir w ⍝ assume max 1 result only
    ∇

    ∇ s←{dropLastFS}fixFsep w;i
      :Access Shared Private
⍝ Validate and Replace / or \ by proper OS file delimiter
      i←i/⍳⍴i←'/\'∊⍨s←w ⋄ s[i]←FS  ⍝ fix Folder separator
      ⎕SIGNAL 22⍴⍨WIN∧':\'>.=1↓3↑s ⍝ cannot be relative path if drive specified
      :If (1↑s)≡,'⍵'                   ⍝ turn ⍵ or ⍵/ into current workspace path
          :If FS≠1↑1↓s
              s←('⍵',FS),1↓s
          :EndIf
          :If 'CLEAR WS'≡8↑ws←⎕WSID   ⍝ if in clear ws, use current directory
              s[⎕IO]←'.'
          :ElseIf isRelPath ws          ⍝ if a ⎕WSID is a Windows relative path
              s←'.',FS,(pathOf ws),FS,2↓s
          :Else
              s←(pathOf ws),1↓s,(1=⍴s)/FS
          :EndIf
      :EndIf
      :If (2↑s,FS)≡'.',FS          ⍝ turn ./ into same path as current directory
          s←((-(¯1↑i)∊'/\')↓i←CD''),1↓s,(1=⍴s)/FS
      :EndIf
      :If 0=⎕NC'dropLastFS'
      :OrIf dropLastFS
          s↓⍨←-FS=¯1↑s
      :EndIf
    ∇

    ∇ b←isRelPath path;s
      :Access Shared Private
      b←~'/\'∊⍨(⎕IO+2×WIN∧':'∊s)⊃s←3↑path
    ∇

    ∇ r←pathOf filename
      :Access Shared Private
      r←{((1∊b)×-1+⊥⍨~b←⍵∊'/\')↓⍵}filename
    ∇

    ∇ {r}←makeDir path;D
      :Access Shared Private
⍝ Create new directories, ignoring messages if any
      →(⍴r←fixFsep path)↓0
      r←⎕SH'mkdir ',((~WIN)/'-p '),D,r,D←WWIN/'"'
    ∇

    ∇ {old}←CD new;get;set;size;t                             ⍝ Change directory.
⍝ Change/report Directory
      :If ~WIN
          old←,⊃⎕SH'pwd' ⍝ ignore set for the nonce
      :Else
          t←'A*'[1+80=⎕DR'']
          'get'⎕NA'U Kernel32|GetCurrentDirectory',t,' U >0T' ⍝ Associate Get function.
          'set'⎕NA'U Kernel32|SetCurrentDirectory',t,' <0T'   ⍝     "     Set    "
          (size old)←get 260 260                              ⍝ Get current directory.
          :If ×⍴,new                                          ⍝ If target directory.
              ⎕SIGNAL(set⊂new)↓11                             ⍝ Domain error if fails.
          :EndIf                                                ⍝ Old dir is shy rslt.
      :EndIf
    ∇

    ∇ {r}←a fileCopy w
      :Access Shared Private
      (a w)←{q,f,q←'"'/⍨WIN>'"'∊f←fixFsep ⍵}¨a w
      r←⎕SH((⎕IO+WIN)⊃'cp ' 'copy /y '),a,' ',w
    ∇
    ∇ {r}←fileErase w;q
      :Access Shared Private
      r←⎕SH((⎕IO+WIN)⊃'rm ' 'erase '),q,(fixFsep w),q←(WIN>'"'∊w)/'"'
    ∇
    ∇ {r}←a fileMove w;to
      :Access Shared Private
      (a w)←{q,f,q←'"'/⍨WIN>'"'∊f←fixFsep ⍵}¨a w
      to←2⊃FS splitLast w ⍝ move to same folder
      r←⎕SH((⎕IO+WIN)⊃'mv ' 'rename '),a,' ',to,⍨WIN/'"'
    ∇

    ∇ files←{parms}Dir x;switches;dfa;⎕IO
⍝ List directory using DOS DIR command
⍝ parms may be any of
⍝ dD: list dirs only
⍝ fF: list files only
⍝ aA: list all data (timestamp, size, name)
      :Access Shared Private
      'parms'Default'' ⋄ ⎕IO←1
      files←WIN NtDirX{⍺:⍺⍺ ⍵ ⋄ ⍵⍵ ⍵}DirU x
      :If ∨/2↑dfa←∨/3 2⍴'dDfFaA'∊parms
          files←((⎕IO⊃files)=1↑dfa)∘⌿¨files
      :EndIf
      :If 0=3⊃dfa
          files←4⊃files
      :EndIf
    ∇

    ∇ files←DirU x;list;cut;b;dir;⎕IO;⎕ML;none
⍝ Unix ls: drop 1st line?, cut into sections
      ⎕IO ⎕ML←1 0
      files←0⍴¨0 '' 0 '' ⋄ none←1
      :Trap 11          ⍝ drop .
          none←0∊⍴list←' ',↑{('total'≡5⍴↑⍵)↓⍵}⎕SH'ls -ld ',x,' 2>/dev/null'
      :EndTrap
      →none⍴0
      b←cut⍲1⌽cut←∧⌿' '=↑list
      dir←list[;2]='d'
      files←1⊃⌽list←0 1∘↓¨⊃⊂/b∘/¨cut list
      files←{(-⊥⍨' '=⍵)↓⍵}¨↓(⌽∧\FS≠⌽1⌷[1]files⍪' ')/files
⍝ 1st section should be permissions, last is name
      files←dir(↓⍕5↓¯1↓list)(2⊃⎕VFI,' ',5⊃list),⊂files
    ∇

    ∇ rslt←{att}NtDirX path;handle;next;ok;attrs;⎕IO;FindFirstFileA;FindNextFileA;FindClose;FileTimeToLocalFileTime;FileTimeToSystemTime;GetLastError;max;keep;mask
⍝ Return NT directory information
      :Access Shared Private
      ⎕IO←0 ⋄ rslt←4⍴⊂'' ⋄ max←2*32
      keep←(⍳32)~30 29 ⍝ all but Hidden & System
      :If 0≠⎕NC'att'
          keep,←('HS'∊att)/30 29
      :EndIf
      mask←(⍳32)∊keep
      FindDefine
      handle next←FindFirstFile path
      →handle↓0 ⍝ file not found
      rslt←,⊂next
      :While 1=0⊃ok next←FindNextFile handle
          rslt,←⊂next
      :EndWhile
      :If 0 18∨.≠ok next
          ('ntdir error:',⍕next)⎕SIGNAL 11
      :EndIf
      ok←FindClose handle
      rslt←↓[0]1 0 0 1 1 0 1 0/⊃rslt       ⍝ bin the unwanted elements
      (0⊃rslt)←(attrs←(32⍴2)⊤0⊃rslt)[27;]  ⍝ Get attributes into bits, note dirs
      rslt←(mask∧.≥attrs)∘/¨rslt           ⍝ discard hidden/system unless wanted
      rslt[1]←Filetime_to_TS¨¨rslt[1]      ⍝ put times into ⎕ts format
      (2⊃rslt)←max⊥max|⍉⊃2⊃rslt            ⍝ combine size elements
      rslt←(~(3⊃rslt)∊1 2⍴¨'.')∘⌿¨rslt     ⍝ remove . & ..
    ∇

    ∇ rslt←Filetime_to_TS filetime;⎕IO
      :Access Shared Private
      :If 1≠0⊃rslt←FileTimeToLocalFileTime filetime(⎕IO←0)
      :OrIf 1≠0⊃rslt←FileTimeToSystemTime(1⊃rslt)0
          rslt←0 0                   ⍝ if either call failed then zero the time elements
      :EndIf
      rslt←1 1 0 1 1 1 1 1/1⊃rslt    ⍝ remove day of week
    ∇

    ∇ FindDefine;WIN32_FIND_DATA;T
      :Access Shared Private
      T←'A*'[80∊⎕DR'']
      WIN32_FIND_DATA←'{I4 {I4 I4} {I4 I4} {I4 I4} {U4 U4} {I4 I4} T[260] T[14]}'
      'FindFirstFileA'⎕NA'I4 kernel32.C32|FindFirstFile',T,' <0T >',WIN32_FIND_DATA
      'FindNextFileA'⎕NA'U4 kernel32.C32|FindNextFile',T,' I4 >',WIN32_FIND_DATA
      ⎕NA'kernel32.C32|FindClose I4'
      ⎕NA'I4 kernel32.C32|FileTimeToLocalFileTime <{I4 I4} >{I4 I4}'
      ⎕NA'I4 kernel32.C32|FileTimeToSystemTime <{I4 I4} >{I2 I2 I2 I2 I2 I2 I2 I2}'
      ⎕NA'I4 kernel32.C32∣GetLastError'
    ∇

    ∇ rslt←FindFirstFile name;⎕IO
      :Access Shared Private
      rslt←FindFirstFileA name(⎕IO←0)
      :If ¯1=0⊃rslt                   ⍝ INVALID_HANDLE_VALUE
          rslt←0 GetLastError
      :Else
          (1 6⊃rslt)trimAt←1↑⎕AV   ⍝ shorten the file name at the null delimiter
          (1 7⊃rslt)trimAt←1↑⎕AV   ⍝ and for the alternate name
      :EndIf
    ∇

    ∇ rslt←FindNextFile handle;⎕IO
      :Access Shared Private
      rslt←FindNextFileA handle(⎕IO←0)
      :If 1≠0⊃rslt
          rslt←0 GetLastError
      :Else
          (1 6⊃rslt)trimAt←1↑⎕AV   ⍝ shorten the filename
          (1 7⊃rslt)trimAt←1↑⎕AV   ⍝ shorten the alternate name
      :EndIf
    ∇

    ∇ r←{remove}GetUnicodeFile file;tn;nc;b;clean
⍝ Read a Unicode (UTF-8 or even UCS-2) file
⍝ This version allows excluding specific 1-byte characters before the translation
⍝ This prevents TRANSLATION errors in classic interpreters
     
      nc←⍴r←256|⎕NREAD tn 83,2↑⎕NSIZE tn←file ⎕NTIE 0 ⋄ ⎕NUNTIE tn
      clean←{⍵}
      :If 2=⎕NC'remove'
          clean←~∘remove
      :EndIf
     
      :If 239 187 191≡3↑r ⍝ UTF-8 header
          r←'UTF-8'⎕UCS clean 3↓r
      :ElseIf ∨/b←(2 2⍴254 255 255)∧.=2↑r ⍝ UCS2
          r←⎕UCS clean 1↓(256*b)+.×⍨(2,⍨nc÷2)⍴r
      :Else ⍝ assume UTF-8 and trap any error
          :Trap 0
              r←'UTF-8'⎕UCS clean r
          :Else
              ⎕SIGNAL ⎕EN
          :EndTrap
      :EndIf
    ∇

    ∇ text PutUTF8File file;tn;⎕IO
⍝ Write Unicode text to new file on disk
      :Trap 22
          file ⎕NERASE file ⎕NTIE ⎕IO←0
      :EndTrap
      ¯17 ¯69 ¯65 ⎕NAPPEND tn←file ⎕NCREATE 0 ⍝ write unsigned UTF-8 header
      (⍳256)⎕NXLATE tn ⍝ for Classic
    ⍝ We cannot use ⎕UCS because of translation errors in Classic
      text←⎕AV['UTF-8'⎕UCS text]
      text ⎕NAPPEND tn 82
      ⎕NUNTIE tn
    ∇


⍝ =====  Registry functions

    ∇ r←SALTsetFile ⍝ file used under Unix for SALT settings
      :If 0=⎕NC'SALTSettingsFile'
          SALTSettingsFile←r←(⎕IO⊃⎕SH'echo ~',⎕AN),'/.dyalog/SALT.settings'
      :Else  ⍝ grab cached value
          r←SALTSettingsFile
      :EndIf
    ∇

    ∇ r←getEnvir w;t
      :Access Shared Private
⍝ SALT is a special env var that is assumed to be $DYALOG/SALT if not present
      :If (w≡'SALT')∧t←0=⍴r←2 ⎕NQ'.' 'GetEnvironment'w
          r←{⍵↓⍨-FS=¯1↑⍵}2 ⎕NQ'.' 'GetEnvironment' 'DYALOG'
          r←(¯4×('bin',FS)≡¯4↑r)↓r
          r←r,FS,'SALT' ⍝ default
      :ElseIf WIN<t
      :AndIf (,'0')≡⎕IO⊃⎕SH'test -f ',SALTsetFile,'; echo $?' ⍝ suggested by AS
⍝ Read setting from the INI files
          r←⎕NREAD t,(⎕DR''),2↑⎕NSIZE t←SALTsetFile ⎕NTIE 0 ⋄ ⎕NUNTIE t
          (r t)←↓⍉⊃' 'splitOn1st⍨¨r splitOn ⎕AV[3]
          r←rlb(r⍳⊂w)⊃t,⊂''
      :EndIf
    ∇

    ∇ value←regSetting arg;name;canbeempty;v;s
⍝ Return setting from the registry/environment
⍝ arg is: [1] Key, [2] default if not present or if empty, and [3] can it be empty
      arg←,condEncl arg
      (name value canbeempty)←arg,(⍴arg)↓'' '' 1
      :Access Shared Private
⍝ We must take into account the fact that SALT's default folder may have been
⍝ superseded by an environment variable:
      :If v≡(s←⍴v←SALTFOLDER)↑value
      :AndIf 0<⍴v←getEnvir'SALT'
          value←v,s↓value  ⍝ default value
      :EndIf
      :Trap 0
          :If canbeempty∨0<⍴v←getEnvir(WIN/'SALT\'),name
              value←v
          :EndIf
      :EndTrap
    ∇

    ∇ name saveSettings value;key;txt;v;n;L;tie;t;i;file
⍝ Save name with value in registry
      :Access Shared Private
      key←3⊃SettingsTable[SettingsTable[;1]⍳⊂name;]
      :If WIN
          regPutString key value
      :Else ⍝ must be Unix world
          file←SALTsetFile
          :Trap 22 ⍝ read setting from the settings files
              txt←⎕NREAD tie,(⎕DR''),2↑⎕NSIZE tie←file ⎕NTIE 0
              n←⍴txt←txt splitOn L←⎕AV[3]
              t←1⊃¨txt splitOn1st¨' '
              :If n<i←t⍳⊂key
                  txt,←0
              :EndIf
              txt[i]←⊂key,' ',⍕value
              txt←1↓⊃,/L,¨txt
          :Else
              :Trap 22
                  tie←file ⎕NCREATE 0
              :Else
                  {}makeDir pathOf file ⋄ tie←file ⎕NCREATE 0
              :EndTrap
              txt←key,' ',⍕value
          :EndTrap
          0 ⎕NRESIZE tie ⋄ txt ⎕NAPPEND tie ⋄ ⎕NUNTIE tie
      :EndIf
    ∇

⍝ The following functions were modified from
⍝ APL Team Ltd ⋄ http://www.aplteam.de ⋄ mailto:kai@aplteam.com

    ∇ regClose HANDLE;RegCloseKey
      :Access Shared Private
      ⎕NA'U ADVAPI32.dll.C32|RegCloseKey U'
      {}RegCloseKey HANDLE
    ∇

    ∇ HANDLE←regGetHandle KEY;HKEY;rck;KEY_ALL_ACCESS;T
      :Access Shared Private
      HKEY←2147483649             ⍝ 'HKEY_CURRENT_USER' HEX 0x80000001
      KEY_ALL_ACCESS←983103       ⍝ HEX 0xF003F
⍝ The next line covers the case for reg & Unicode versions
      T←'A*'[⎕IO+80∊⎕DR'']
      'rck'⎕NA'I ADVAPI32.dll.C32|RegCreateKeyEx',T,' U <0T I <0T I I I >U >U'
      HANDLE←⊃2⊃rck HKEY KEY 0 '' 0 KEY_ALL_ACCESS 0 0 0
    ∇

    ∇ regPutString(SUBKEY STRING);set;REG_SZ;Path;HANDLE;T;uni
⍝ Stores the value of a Registry SUBKEY
      :Access Shared Private
      Path←(getEnvir'inifile'),'\SALT'
      HANDLE←regGetHandle Path
      STRING←,STRING
      REG_SZ←1 ⍝ String data type
      T←'A*'[uni←1+80∊⎕DR'']
      'set'⎕NA'I ADVAPI32.dll.C32|RegSetValueEx',T,' U <0T I I <0T I4'
      {}set HANDLE SUBKEY 0 REG_SZ STRING(uni×1+⊃⍴STRING)
      regClose HANDLE
    ∇

:EndNameSpace ⍝ SALTUtils  $Revision$ $Date$