﻿:Class FnDiff               
⍝ Implement FnDiff as a Spice Command. Based on fndiff from ws dfns.
    
    ∇ r←List
      :Access Shared Public
      r←⎕NS''
      r.Name←'fndiff'
      r.Desc←'Compares functions in any namespace'
      r.Parse←'2'
      r.Group←'wsutils'
    ∇
    
    ∇ r←Run(Cmd Args);a
      :Access Shared Public
      a←Args.Arguments
      r←fndiff{((~'.'∊⍵)/'.',⍨⍕##.THIS),⍵}¨a ⍝ ensure we get the fn from # if not specified
    ∇
    
    ∇ r←lev Help Cmd
      :Access Shared Public
      r←r[⍳2⌈(×lev)×1↑⍴r←3 1↓⎕CR'Help';]
⍝Show the different lines in two functions
⍝Syntax: fndiff "[ws1] fn1" "[ws2] fn2"
⍝
⍝ Given
⍝       ⎕cr¨'Fa' 'Fb'
⍝  Fa           Fb
⍝  dif1a        dif1b
⍝  same1        same1
⍝  insertedFa   same2
⍝  same2        dif2b
⍝  dif2a        insertedFb
⍝  same3        same3
⍝
⍝       ]fndiff  Fa  Fb
⍝·Fa ·   ·   ·   ·   ·   │·Fb ·   ·   ·   ·   ·
⍝·dif1a  ·   ·   ·   ·   │·dif1b  ·   ·   ·   ·
⍝·insertedFa ·   ·   ·   │·dif2b  ·   ·   ·   ·
⍝·dif2a  ·   ·   ·   ·   │·insertedFb ·   ·   ·
    ∇

      fndiff←{⎕ML←0                               ⍝ Defined function differences.
          ⍺←'' ⋄ {                                ⍝ diffs between two functions.
              mats←↑¨0 1{↑~/⍺⌽⍵}¨⊂⍵               ⍝ line differences.
              cols←⌊0.5×⎕PW-1                     ⍝ cols per half-screen.
              size←(⌈/⊃∘⍴¨mats)cols               ⍝ padding size.
              ↑{⍺,'│',⍵}/{                        ⍝ aligning dots.
                  mask←,(⍵=' ')∧(⍴⍵)⍴cols⍴4↑1     ⍝ blanks every 4 cols,
                  mat←⍵ ⋄ {mat}(mask/,mat)←'·'    ⍝ inject into char mat.
              }¨size∘↑¨mats                       ⍝ split screen output.
          }(' ⍝'∊⍺)∘{                             ⍝ trim white space exclusions.
              ⍺≡0 0:⍵                             ⍝ neither excluded: continue.
              cm←{/∘⍵∧\(⍵≠'⍝')∨≠\''''=⍵}          ⍝ remove unquoted comments.
              sp←{(∨\' '≠⍵)/⍵}                    ⍝ remove leading blanks.
              ⍺{⍵~(⊃⍺)/⊂''}⍺∘{                    ⍝ with/out blank lines.
                  ⍺≡1 1:⌽sp⌽sp cm ⍵               ⍝ remove blanks and comments, or
                  ⍺≡0 1:⌽sp⌽cm ⍵                  ⍝ ... comments & trailing blanks.
                  code←cm ⍵                       ⍝ uncommented code.
                  cmnt←(⍴code)↓⍵                  ⍝ comment from code.
                  (⌽sp⌽sp code),cmnt              ⍝ before and after blanks removed.
              }¨⍵                                 ⍝ ... from each function line.
          }∘{                                     ⍝ copy in function if remote.
              split←{(1 0=⊂⌽∨\' '=⌽⍵)/¨⊂⍵}        ⍝ split vector at rightmost blank.
              trim←{(∨\' '≠⍵)/⍵}{⌽⍺⍺⌽⍺⍺ ⍵}        ⍝ remove leading/trailing blanks.
              wsid fnid←trim¨split trim ⍵         ⍝ wsid / fnid.
              ''≡wsid~' ':⎕NR fnid                ⍝ function in current ws.
              base←{(⌽∧\⌽⍵≠'.')/⍵}fnid            ⍝ base name of function.
              copy←'fnid ⍺.⎕cy wsid'              ⍝ command: copy remote function.
              srce←'⍺.⎕nr base'                   ⍝ command: return function lines.
              0::('Can''t copy: ',⍵)⎕SIGNAL ⎕EN   ⍝ abort if copy fails.
              (⎕NS''){⍎⍵}copy,'⋄',srce            ⍝ copy and return function lines.
          }¨⍵                                     ⍝ ... for each function.
      }
    
:EndClass