﻿:namespace  inout

    ⎕IO←1 ⋄ ⎕ML←0

    ⍝ There are 2 possibilities:
    ⍝ 1. we are dealing with ATF files (extension .ATF)     : in / out
    ⍝ 2. we are dealing with Dyalog's (own) eXtended format : inx/ outx

    Cmds←'in' 'out' 'inx' 'outx'

    APLs←'APL2PC' 'APLX' 'APLC'     'APL2MF' 'DYALOG' 'APLPLUS' 'VSAPL' 'STSCMF'
    Surr←'A2K'    'APX'  ']in only' 'AP2'    'DYW'    'A2K'     'AP2'   'A2K'   

    ⍝ We could have created only 2 cmds (in & out) to avoid having another set of 2 but because
    ⍝ their switches are different it was written this way.

    ∇ r←List
      :Access Shared Public
      r←⎕NS¨(⍴Cmds)⍴⊂''
      r.Name←Cmds
      r[1].Desc←'Define an ATF file''s objects in the ws' ⍝ descriptions
      r[2].Desc←'Create an ATF file from the current ws'
      r[3].Desc←'Define in the ws the objects of an extended transfer file'
      r[4].Desc←'Create an extended transfer file from the current ws'
      r.Group←⊂'Transfer'                                 ⍝ same group
     ⍝ Parsing rules for each:
      r[1].Parse←'1 -fliplu -apl='
      r[2].Parse←'1 -obj='
      r[3].Parse←'1 -apl= -file[=] -list -noam -obj= -q -range= -replace -trans[=]0 1 2'
      r[4].Parse←'1 -obj= -file[=] -range= -lock= -q'
    ∇

    ∇ r←Run(Cmd Args);arg;b;q;nars
      :Access Shared Public
      arg←1⊃Args.Arguments
      :If nars←'x'∊Cmd
         ⍝ xfr assumes the extension to be specific. We can guess if none supplied:
          :If ~'.'∊arg
          :AndIf 1=⍴q←'f'⎕SE.SALTUtils.Dir'.*',⍨arg~'"' ⍝ no doubt?
              arg{⍺,(∨\'.'=⍵)/⍵}←1⊃q
          :EndIf
          ##.⎕CY'xfrcode'
          :If 'inx'≡Cmd ⋄ nars←(∨/Args.apl∘≡¨'nars' 'NARS')∨'.nars'≡¯5↑arg~'"' ⋄ :EndIf
      :Else
          ⎕FIX ⎕SE.SALT.Load'lib\atfin -source'
      :EndIf
     
      :If nars<∨/b←'''" -'∊arg
          ⍝ The fns below will reparse the line so we need to surround the arg properly with quotes
          arg←q,((1+q=arg)/arg),q←''''
      :End
      :Select Cmd
      :Case 1⊃Cmds
          ##.THIS ⍙⍙.ATFIN arg,Args.Propagate'fliplu apl'
      :Case 2⊃Cmds
          ##.THIS ⍙⍙.ATFOUT arg,Args.Propagate'obj'
      :Case 3⊃Cmds
          :If nars
              ##.THIS ##.xfr.nars.makeWS arg
          :Else
              :If ∨/b←APLs∊⊂Args.apl ⋄ Args.apl←(b⍳1)⊃Surr ⋄ :EndIf
              ##.THIS ##.xfr.∆xfrfrom arg,Args.Propagate'apl file list noam obj q range replace trans'
          :EndIf
      :Case 4⊃Cmds
          ##.xfr.∆xfrto arg,Args.Propagate'file range lock obj q'
      :EndSelect
      r←0 0⍴''
    ∇

    ∇ r←lev Help Cmd;o;i;ix;ox
      :Access Shared Public
      :If Cmd≡'in'
          i←⊂'Bring objects defined in an ATF file.'
          i,←⊂'The source APL can be specified through the -apl switch. The acceptable APLs are:'
          i,←⊂(⍕APLs)
          i,←'' 'If the source APL was using underscored letters you can replace them by lowercase'
          i,←(⊂'(switch them) by using the -fliplu switch.'),(lev=0)/⊂'See ]??in for details'
          r←↑i ⋄ →lev↓0
          i,←'' 'ATF files (APL Transfer Format) are files used historically to transfer APL code.'
          i,←⊂'They were introduced in the 80s and are sometimes used to recover code from other'
          i,←⊂'APL vendors. They don''t always follow the same format and can be difficult to deal with.'
          r←↑i,'' 'Example: bring in an ATF file generated by APL+' '      ]in \tmp\aplp'
      :ElseIf Cmd≡'out'
          o←⊂'Write objects, vars and fns only (default all) in current ns, to an ATF file.'
          o,←⊂'You can specify which objects to be written out with the -obj switch'
          r←↑o ⋄ →lev↓0
          r←↑o,'' 'Example:' '      ]out \tmp\aplp'
      :ElseIf Cmd≡'inx'
          ix←⊂'Bring objects defined in an extended transfer file or a .nars file.'
          ix,←⊂'Several switches allow to fine tune the process.',(lev=0)/' See ]??inx for details.'
          r←↑ix ⋄ →lev↓0
          ix,←'' 'Examples:' 'Retrieve all objects from an APL+Win transfer file:' '      ]inx \tmp\myws.xsw'
          ix,←'' 'Retrieve some objects from a Dyalog transfer file:' '      ]inx \tmp\myws.xdw -obj=A BC def'
          ix,←'' 'Retrieve all objects in a NARS workspace:' '      ]inx \full\path\to\extws.ws.nars'
          ix,←⊂'Note that you cannot retrieve individual objects from a NARS ws nor perform translation (yet).'
          ix,←'' 'For a complete description of the code do' '      ]???inx'
          r←↑ix ⋄ →0/⍨lev=1
          'xfr.∆describe'⎕CY'xfrcode'
          r←∆describe
      :ElseIf Cmd≡'outx'
          ox←⊂'Write objects (default all in current ws) to an extended transfer file.'
          ox,←⊂'Specific objects can be selected with the -obj switch.',(lev=0)/' See ]??outx for details.'
          r←↑ox ⋄ →lev↓0
          ox,←'' 'Example:' '      ]outx \tmp\myws -obj=A BC def'
          ox,←'' 'To retrieve the objects in another APL, )LOAD the xfrpc ws on that APL and do' '      ∆xfrfrom ''\tmp\myws.xdw'''
          ox,←'' 'To retrieve the objects in (possibly another version of) Dyalog APL do' '      ]inx \tmp\myws.xdw'
          ox,←'' 'For a complete description of the code do' '      ]???outx'
          r←↑ox ⋄ →0/⍨lev=1
          'xfr.∆describe'⎕CY'xfrcode'
          r←∆describe
      :EndIf
    ∇

:Endnamespace