﻿:namespace wsloc

⍝ Locate strings in the workspace. See 'Describe' for details,

    (⎕IO ⎕ML ⎕PP)←0 2 16                                     ⋄ Version←6.19

    ∇ r←List
      :Access Shared Public
      r←⎕NS¨1⍴⊂⍬
      r.Name←⊂'wslocate'
      r.Group←⊂'wsutils'
      r[0].Desc←'locate and replace strings in the workspace'
      r[0].Parse←''
    ∇

    ∇ r←Run(Cmd Args)
      :Access Shared Public
      r←##.THIS SS Args,(0≠⎕SE.SALTUtils.DEBUG)/' -DEBUG'
    ∇

    ∇ r←{details}Help Cmd
      :Access Shared Public
      :If 0=⎕NC'details' ⋄ details←0 ⋄ :EndIf
      :If details>0 ⋄ r←Describe
      :Else ⋄ r←Synopsis
      :EndIf
    ∇

    (CR NL)←⎕av[3 2] ⋄ C←CR

    Switches←_←'-OBJECTS= -CLASS∊234VFOvfo -RECURSIVE[=] -EXCLUDE∊BbTtCcLlGg -APL -INSENSITIVE -NUMERIC -SHOW ' ⋄ _,←C
    Switches,←_,←'-SYNTACTIC[∊]LlRr -PATTERN[=] -DEBUG -NOBANNER -REPLACE -RETURN[=]count details each text'

    _←'Switches are:' ,C,_,C,C
    _,←'This command will search all the objects specified for the strings in the right argument.'
    _,←C,'By default the objects searched are functions (and operators if they exist). '
    _,←C,'The strings are separated by spaces in the right argument.'
    _,←C,'Each line containing a searched string is displayed and a caret (∧) points to the place '
    _,←C,'where a match is found. If the string searched is a regular expression the entire match '
    _,←C,'is underlined with carets.',C
    Synopsis←_
                               
    ∇ r←GenHelp
      r←↑,/CR,¨2↓2↓¨⎕NR'GenHelp'
⍝ In its simplest form doing
⍝     ]wsloc abc
⍝ will search for the string 'abc' in all functions in the active namespace.
⍝
⍝ The behaviour of WSLOC can be modified with the use of switches (options).
⍝
⍝ The possible switches are:
⍝
⍝ -EXCLUDE= C T B L G
⍝  B    exclude the object's Body (code)
⍝  C    exclude the object's Comments
⍝  T    exclude the object's Text within quotes
⍝  L    exclude if localized (if there is a match in the first row of a Trad fn)
⍝  G    exclude if NOT localized (if NO match in the first row)
⍝
⍝ -APL  consider strings as APL code after ⍎, ⎕trap, etc.
⍝
⍝ -OBJECTS= list of objects to look into
⍝
⍝ -CLASS=   list of classes (e.g. 3=fns) to look into. Class 9 implies RECURSIVE.
⍝  Note that only one of CLASS or OBJECTS can be specified.
⍝
⍝ -INSENSITIVE  insensitive (case) search
⍝
⍝ -NUMERIC      include numeric (non char) variables in search.
⍝ If the search strings only include numbers then the numeric
⍝ variables will be automatically searched.
⍝
⍝ -PATTERN the strings are patterns (see further below).
⍝ The pattern matching code may come from .Net in multiline or single line mode
⍝ if ⎕S is unavailable. Minimum .Net 2.0 must be installed.
⍝ .Net will be used if ⎕S is unavailable otherwise it can be selected.
⍝
⍝ -RETURN [=EACH|DETAILS|TEXT|COUNT]
⍝ The function returns an empty result unless this switch is used. The shape of
⍝ the result is 0 followed by the number of hits, e.g. 0 47 for 47 total hits.
⍝
⍝ If COUNT is supplied as value to RETURN then a matrix of integers
⍝ is returned, one row per string, one column per object searched.
⍝ Each number represents the number of hits of a string for a given object.
⍝ To know to which string/object the number belongs to an extra row on top
⍝ shows the object and an extra column to the left shows the string searched.
⍝
⍝ If EACH is supplied as argument to RETURN then a vector of enclosures is returned,
⍝ each element being [1] the object name followed by [2] a 3D integer structure.
⍝ In that 3D structure each plane represents a hit.
⍝ The first row describes the hit with 5 numbers:
⍝ 1. position in text
⍝ 2. length of hit
⍝ 3. searched string number
⍝ 4. row
⍝ 5. column
⍝
⍝ The other rows will only exist if a pattern with substring was used,
⍝ each row belonging to a substring. The numbers have the same meaning.
⍝
⍝ If DETAILS is supplied as value to RETURN then the same info as EACH
⍝ is supplied plus 1 more column for each row: the text found.
⍝ NOTE: the text of the line where it is found is not returned. It is easy enough
⍝ to get that information as part of the regular expression itself.
⍝
⍝ If RETURN is used alone (without a value) or if TEXT is supplied as value to
⍝ it then the entire output is returned as a string.
⍝ This may cause WSFULLs if the output is large.
⍝
⍝ -RECURSIVE[=]  search Namespaces recursively from the namespace specified
⍝                (default current ns)
⍝
⍝ -NOBANNER      do not display banner message
⍝
⍝ -SYNTACTIC [=L R]
⍝ Syntactic search meaning not preceded (or followed) by
⍝ a letter, digit, ⎕, ∆, ⍙, _ or ¯ characters.
⍝ If the =L value is used LEFT  syntactic search is performed.
⍝ If the =R value is used RIGHT syntactic search is performed.
⍝ If both are used it is the same as using SYNTACTIC without a value.
⍝
⍝ -SHOW will force a display of the hits. The display won't normally be shown
⍝ if RETURN is used.
⍝
⍝
⍝ >>> RESULT
⍝
⍝ The command will, by default, display the lines where the strings are
⍝ found with a caret (∧) below. It then returns an empty result. It
⍝ can also return an integer table with as many columns as objects
⍝ were searched and as many rows as strings were looked for with a row on
⍝ top with object full path names and one column to the left for the string
⍝ searched. For example, if all 20 functions of a workspace were searched
⍝ for 3 strings then a 4 by 21 table of the number of hits per fn/string
⍝ with names would be returned. See section on RETURN above.
⍝
⍝
⍝ >>> EXAMPLES
⍝
⍝     ]wsloc abc     ⍝ search 'abc' everywhere
⍝     ]wsloc abc def ⍝ search 'abc' and 'def' everywhere
⍝     ]wsloc abc   -SYNTACTIC=R              ⍝ search names ending in '...abc'
⍝     ]wsloc abc f -SYN -EXCLUDE=TC -OBJ=x y ⍝ syntactic search of 'abc' and 'f'
⍝                    in objects 'x' and 'y' not within Text or Comments
⍝
⍝ In the following example we have a file of ⎕CRs of functions and
⍝ we want to know where variables 'x' and 'yz' are used AND not
⍝ localized in components starting at 11 (file is tied to 8):
⍝
⍝     ]wsloc x yz -EX=LCT -APL -SYN -OBJ=8 11  ⍝ look for 'x' and 'yz' as exact
⍝ names (SYN) in file tied to 8 starting at component 11. Don't look in
⍝ comments or text (EX=CT) except where ⍎ is concerned (APL) and
⍝ exclude where localized (EX=L).
⍝
⍝     ]wsloc x y z -OB=F G -INS -SYN -RETURN ⍝ return a 4×3 of the number of hits in
⍝ each of <F> and <G> for lower and uppercase names 'x', 'y' and 'z'.
⍝ The table will include a extra column for the strings and extra row for the names.
⍝
⍝ To include space and - (the switch delimiter) in the search surround the text with
⍝ double quotes or quotes:
⍝
⍝     ]wsloc 'ABC← -/¨A B C'
⍝ will look for where ABC is assigned the -/ of each of A, B and C. Otherwise
⍝ the cmd will report an error on the /¨A switch.
⍝
⍝ To replace syntactic 'abc' by 'xyz' in all the code of each fn in the current namespace:
⍝     ]wsloc abc xyz -SYN -REPLACE -EXCLUDE=TC'
⍝
⍝
⍝ >>> PATTERN MATCHING
⍝
⍝ Pattern matching is possible to search and replace. Two ways are possible, either using
⍝ ⎕S or by using .NET (Windows only). V13 will use ⎕S by default.
⍝
⍝ For example, we wish to find where files are tied (or created), that is, find
⍝ out which strings contain a filename. In regex terms: find strings that are "names separated
⍝ by / or \ possibly prefixed by //name or LETTER: and a /".
⍝ To do that we'd enter something like this:
⍝
⍝     ]wsloc "'(//\w+/|[a-zA-Z]:/)?([/\\]?\w+)+'"  -PATTERN -EX=CB
⍝ to look for a filename within text of all fns (EX=CB excludes body and comments).
⍝ Because we want to include the quotes in the search we must surround the expression with quotes
⍝ and insert the whole thing inside double quotes.
⍝
⍝ To simplify search of names and numbers, the pseudo patterns ⍺ and ⍵ can be used.
⍝ For example to see where 3 names are assigned 3 numbers you could do:
⍝
⍝     ]wsloc '⍺ ⍺ ⍺←⍵ ⍵ ⍵' -PATTERN
⍝
⍝ When displaying hits the caret (∧) is displayed under the whole match, not just the
⍝ beginning as in non-regex mode. Furthermore, if the match spans several lines, (can
⍝ only happen in multiline mode) only the first line is displayed and the caret extends
⍝ 1 further to denote the fact. Ex: Given <F>:
⍝     ∇ F A
⍝ [1]  ⎕←A
⍝     ∇
⍝
⍝ Doing  "]wsloc A.*A -P=single"  would find a match spanning from the 'A' in the header
⍝ to the 'A' on line 1 and the display would look like this:
⍝
⍝     ∇ #.F (1 found)
⍝ [0]   F A
⍝         ∧∧
⍝ The carets extend past 'A' to denote that the match continues further.
⍝
⍝ To use .Net use -PATTERN=dnet (this is the default under V12).
⍝
⍝
⍝ >>> REPLACEMENT
⍝
⍝ To perform replacement use pairs of search-string replacement-string and use
⍝ the -REPLACE switch. If you use also the -SHOW switch you will also see where
⍝ the changes are made.
⍝
⍝ Pattern matching replacement is also possible. If specific sub-patterns
⍝ are to be reused you must use the backslash followed by their position number
⍝ to denote them. For example, to change the string 'abc(something)123' into
⍝ '123(something)abc' you should do
⍝
⍝     ]wsloc (abc)(.*)(123) \3\2\1 -REPLACE -PATTERN
⍝
⍝ For each match of the 2 strings 'abc' and '123' this will effectively switch
⍝ them around, putting the 3rd string (\3) before the 1st one (\1).
⍝ Each substring must be surrounded by parentheses as usual.
⍝
⍝ >>> LIMITATIONS
⍝
⍝   This will search functions, variables and files. This version
⍝ can also look at namespaces. If an object to search is a namespace it is treated
⍝ as if the "recursive" switch had been set.
⍝
⍝   Excluding locals or globals only makes sense for FULL syntactic search in fns.
⍝ The command will simply look for a match in the first line of a Trad fn and
⍝ either continue or abort depending on the request.
    ∇

    Describe←Synopsis,GenHelp

    Cset←'_abcdefghijklmnopqrstuvwxyz∆ABCDEFGHIJKLMNOPQRSTUVWXYZ⍙ÁÂÃÇÈÊËÌÍÎÏÐÒÓÔÕÙÚÛÝþãìðòõÀÄÅÆÉÑÖØÜßàáâäåæçèéêëíîïñóôöøùúûü0123456789⎕'
    Uset←3 27⍴Cset

    _←0≤⎕NC 256 1⍴⎕AV
    _←'a-zA-Z_',(_⌿⎕AV)~,⎕A,27⍴Cset
   ⍝ ?: is non capture, ?> is atomic grouping, ?<! is negative lookbehind
    AP←'(?:(?<!\s:|::)(?<![⎕0-9',_,'])[',_,'][',_,'0-9]*|⍺⍺|⍺|⍵⍵|⍵)' ⍝ token name (⍺) pattern - no ⎕name
    OP←'¯?(?<![',_,'0-9])(?>\d+\.?\d*|\d*\.?\b\d+)(?>[eE]¯?\d+)?' ⍝ # (⍵) pattern
    OP←'(?:',OP,'(?>[jJ]',OP,')?)' ⍝ complex number

    ⎕ex¨'C_'

      remRef←{names←(b←0<cl←⍺.⎕NC n)/n←⍺.⎕NL ⍵         ⍝ ensure visible
          ~∨/n←9.1=b/cl:names                          ⍝ no namespaces
          (names/⍨~n),⍺{⍵/⍨(⍺≠s)∧⍺=(s←⍺⍎¨⍵).##}n/names ⍝ eliminate refs
      }
      
    Msg←{~0∊⍴⍵:⎕←⍵} ⍝ can be eventually changed to account for ]ucmd call

    ∇ çRz←{LA}SS RA;t1;btc;Rstr;var;id;NOBANNER;SHOW;DEBUG;PATTERN;INSENSITIVE;RETURN;APL;RECURSIVE;REPLACE;EXCLUDE;SYNTACTIC;NUMERIC;CLASS;ç98;ç97;onfile;ç95;ç94;ç93;ç92;ç91;ç90;Strings;inc;⎕PP;⎕ML;CurSpace;t;Path;OBJECTS;nlLA;Pattern;pcr;name;action;lac;SALTed;Typ;out
      :Access shared public
      ç93←'Search String (Find and Replace) for Dyalog V',2⍕Version
     ⍝ This fn will perform String Search and Replace on most APL objects.
     ⍝ It receives a ns as Rarg which contains the Strings to search/replace and the switches OR a string which it parses itself.
     
     ⍝ Syntax:
     ⍝        [namelist] SS 'strings [-switches]'  ('-' is the delimiter to use for switches)
     
     ⍝ NOTE: this version does not replace numbers in numeric vars, nor searches instances.
     
     ⍝ It first defines a series of fns to do the work properly, in this ns, then proceeds to do the work.
     ⍝ If it has to do recursive searches it reuses the fns already defined. Many objects temporarily defined start with a 'ç'.
     
     ⍝ The Larg is the list of objects to search (default all). If it is a ref we search the objects in it.
     
      nlLA←onfile←0 ⋄ id←⍕CurSpace←{0::# ⋄ ⍎⍵}⍕1↑⎕NSI ⍝ NOT '⎕se'⎕WG'curspace' as this can be called from anywhere
      :If 9∊lac←⎕NC'LA' ⋄ id←⍕CurSpace←LA             ⍝ a ref can be presented as such
      :ElseIf 2∊lac                                   ⍝ defined LA?
      :AndIf ~onfile←∨/1 3∊10|⎕DR 1/LA                ⍝ LA could be tie #, etc.
      :AndIf ~nlLA←(,9)≢⎕NC LA ⋄ CurSpace←⍎id←LA      ⍝ a ref could be presented as a string
      :EndIf
      Path←id,'.'
     
     ⍝ Right arg may be a single 1 in which case the left arg is a ns to search
      :If 1≡RA                      ⍝ a single 1 means 'recursive call'
          çRz←çR0                   ⍝ çR0 and ç are global to  this fn and hold common data needed at each (recursize) call
          REDEFINE ç                ⍝ redefine all vars locally
          LA←CurSpace remRef-CLASS  ⍝ the objects to  search
      :Else                         ⍝ proceed to the definition of the work to do
     ⍝ Left arg is a Namelist or tie # and right arg is either a properly formed string with switches...
          :If var←9≠⎕NC'RA'         ⍝ 'var' needed in DEBUG
             ⍝ We require the Parser to be there
              ç←(⎕NEW ⎕SE.Parser(Switches'upper')).Parse RA ⍝ args and switches kept in this global
          :Else ⋄ ç←RA
          :EndIf
     
         ⍝ RECURSIVE may specify a starting point where to recurse from
          :If 1=⍴⍴t←ç.RECURSIVE ⋄ CurSpace←⍎id←t,⍨'#.'/⍨'#'≠1↑t ⋄ Path←id,'.' ⋄ ç.RECURSIVE←1 ⋄ :EndIf
          ç.(EXCLUDE←∨⌿2 5⍴'BTCLGbtclg'∊EXCLUDE)
          ç.(SYNTACTIC←∨⌿3 2⍴'lrLR11'∊⍕SYNTACTIC)
          :If ç.Pattern←Pattern←0≢t←ç.PATTERN
          :OrIf t←ç.APL
              t1←{0::4 ⋄ z←⍎'+⎕r+' ⋄ 2}0 ⍝ 2=⎕S available, 4=use .Net, 6=PCRE DLL (no longer used); + 1 if multiline (the default)
         ⍝ Pattern may be specified as SINGLE (otherwise multi defaulted) or DNET (default or ⎕S if available)
              :If t≡1 ⋄ ç.PATTERN←1+t1 ⍝ w/o a value PATTERN means the default (⎕S for V13+, .Net otherwise)
              :Else ⍝ engine and mode were specified
                  :Trap 701
                      ç90←(⎕NEW ⎕SE.Parser(' dnet single' 'upper allownospace')).Parse t
                  :Else
                      'Unknown value for -pattern='⎕SIGNAL 11
                  :EndTrap
                  ç.PATTERN←4 6 t1[ç90.(DNET 0)⍳1]+~ç90.SINGLE ⍝ 0 2⊤→fn,multi
              :EndIf
          :EndIf
          'Only one of name list, Objects or Class is acceptable'⎕SIGNAL 11 if 1<+/nlLA,0≢¨ç.(OBJECTS CLASS)
          Strings←ç.Arguments
         ⍝ Here all switches have been defined in global 'ç'
          REDEFINE ç
          Msg(NOBANNER<RETURN≡0)/ç93  ⍝ display banner?
          :If DEBUG
             ⍝ Display debugging information
              ç90←((5 5 6/ç90)/'FULL Left Right '),(∨/ç90←<\(∧/ç90),ç90←SYNTACTIC)/'Syntactic '
              ç92←(((ç92⍳'-')-var×⍴ç92)↑ç92←'1 letter options-switches'),' method.'
              ⎕←'fn called in ',ç90,((6-REPLACE×⍴ç92)↑ç92←'SEARCHREPLACE'),' mode using the ',ç92
              ⎕←'Called from ',id
              ⎕←'Number of Strings: ',⍕⍴Strings
              ç90←(∨/EXCLUDE)/('Exclude',(5 5 9 10 8/EXCLUDE)/' BODY TEXT COMMENTS LOCALIZED GLOBALS'),CR
              ç90←ç90,(0≢RETURN)/'The fn returns ',(((⊂RETURN)∊'each' 'EACH')/'individual '),'results',CR
              ç90←ç90,APL/'APL mode ON',CR
              ç90←ç90,INSENSITIVE/'Case insensitive mode ON',CR
              ç90←ç90,NUMERIC/'Numeric vars searched',CR
              ç90←ç90,RECURSIVE/'Recursive mode ON',CR
              t←(1⍴ç92←0 2⊤PATTERN)⊃0 '⎕RX' '.Net' 'PCRE'
              ç90←ç90,(0<PATTERN)/t,' patterns are used ',((~1↓ç92)/', singleline mode '),CR
              ç90←ç90,(~NOBANNER)/'Messages are shown',CR
              ç90←ç90,SHOW/'Hits are displayed',CR
              ⎕←ç90
          :EndIf
          :If Rstr←REPLACE   ⍝ is this a replacement?
              'ERROR: Argument length is odd'⎕SIGNAL(2|t←⍴Strings)⍴11 ⍝ Do we have only pairs then?
              ç.Rstr←Rstr←(t⍴0 1)/Strings
          :EndIf
     
         ⍝ Issue a warning if EXCLUDE locals is used with non full SYNTATIC search:
          Msg((≠/¯2↑EXCLUDE)∧(≠/SYNTACTIC)∧DEBUG≥NOBANNER)/'** Partial SYNTACTIC search used with local exclusion, results can be misleading'
     
         ⍝ We set the Numeric flag if all numeric
          ç.NUMERIC←NUMERIC←NUMERIC∨∧/0⊃⎕VFI⍕Strings
     
         ⍝ Use proper pattern matching fn
          var←'⍷' ⍝ the fn to use to search
          :If APL∨×t←PATTERN ⋄ var←(⍕2|t),'∘PATMATCH',t⊃2/0 'rx' 'dnet' 'dll' ⋄ :EndIf
          çFind←⍎var
     
          CLASS←(t∨0 1 1>∨/t←∨⌿3 3⍴'234VFOvfo'∊CLASS)/2 3 4 ⍝ default to fns/ops
          ç.CLASS←CLASS←CLASS,RECURSIVE/9
          :If onfile⍱nlLA
              (onfile LA)←{⍵≡0:0 ⋄ ∧/⎕IO⊃v←⎕VFI ⍵:1,1↓v ⋄ 0 ⍵}OBJECTS
          :AndIf 0≡LA
             ⍝ do we need ns as well? the recursive method will include them w/o refs
              LA←CurSpace remRef-CLASS ⍝ find objects to search
          :EndIf
     
         ⍝ LA could be tie #, start, last cpts, increment
          :If onfile
              ç92←ç91⌈t←LA,(⍴,LA)↓ç91←0,(2↑⎕FSIZE 1↑LA),1
              ç92[2]←ç91[2]⌊t[2]+1 ⋄ LA←ç92[⍳3]-inc←0,ç92[3],0
          :Else
         ⍝ Ensure left arg is list of objects (VTV)
              :If 1≥|≡LA ⍝ allow enclosed list
                  LA←(t⍲1⌽t←LA∊' ')/LA←,' ',LA ⍝ can be a matrix or space delimited string
                  LA←1↓¨(LA=' ')⊂LA
              :EndIf
          :EndIf
     
         ⍝ The rule for SHOW and RETURN is like this:
         ⍝ If we are in search mode (not replace) then by default we show and return nothing.
         ⍝ If return is not to show (show is 1, 0 or 'text') we DON'T show (if SHOW is set we leave it ON)
         ⍝ If we are in replace mode the conditions are reversed.
         ⍝ In short, we show the hits if requested or a result is wanted and no replacement is made
          ç.SHOW←SHOW←SHOW∨REPLACE<0 1 'text'∊⍨⊂RETURN
          :If REPLACE ⋄ RETURN←0 ⋄ :EndIf ⍝ no other RETURN makes sense in replace mode
     
     
     ⍝ ----------- Define Local functions. Fixed utilities start with a lowercase letter. -------------
     
     ⍝ The shape of the result depends on the request:
     ⍝ by default, the fn returns a (0,totalHits)⍴0. It may return a list of enclosed series
     ⍝ of a fn name followed by a table of integers (3D): ith match, pos, length, row, col.
     ⍝ It can also return the number of hits for each string in each objects (2D int) with titles.
     ⍝ ⍙Rz is initialized accordingly: (exit if nothing to search)
     
          çMrG←0 1∘↓⍣(RETURN≡'count') ⍝ remove strings column of recursive results
          ⎕FX(⊂'r←l GETSTRINGS r'),('details'≡LowerCase RETURN)/⊂'r←r,(⊂¨r[;;0]+⍳¨r[;;1])⌷¨⊂l'
     
          :Select RETURN
          :CaseList 'details' 'each'
              çRz←⍳0                  ⍝ each enclosed result catenated, result is vector
              ç93←⎕FX'r←r çcAr new' 'r←r,⊂2↑new'
          :Case 'count'
              t←(⍴Strings)÷1+REPLACE  ⍝ number of Strings
              çRz←(1+t,0)⍴' ',Strings{⍺/⍨(⍴⍺)⍴1,~⍵}REPLACE    ⍝ each hit # returned for each searched string
              ç93←⎕FX'r←r çcAr (f new x)'('r←r,(⊂f),+⌿new[;0;2]∘.=⍳',⍕t)
          :Case 0
              çRz←0 0⍴0               ⍝ nothing returned (0×ttl hits returned)
              ç93←⎕FX'r←r çcAr (f new x)' 'r←r,(0,1↑⍴new)⍴0' ⍝ final shape will 0xTTL hits
          :Else ⍝ 1 or 'text'
              çRz←'' ⋄ RETURN←1
              ç93←⎕FX'r←r çcAr (x y new)' 'r←r,new'
          :EndSelect
          →(⍴Strings)↓çenD           ⍝ anything to search?
     
          :If INSENSITIVE ⋄ ç.Arguments←Strings←LowerCase¨Strings ⋄ :EndIf ⍝ insensitive to case?
     
          Msg(NOBANNER<0∊∊⍴¨Strings)/'Empty Strings are ignored'
     
          çR0←çRz                    ⍝ initial value (used when recursing)
      :EndIf
     
         ⍝ The "actions" table: this version does not replace numeric values
         ⍝ Each row tells what can be done: add to result, allow search, allow replace
      btc←'CSENI',1,5 2⍴1,REPLACE,1,REPLACE,1,REPLACE,NUMERIC,0 0 0 ⍝ table of actions for each case
     ⍝
     ⍝                                   --------     Main loop     --------
     ⍝
      :While 0<⍴LA
          id←Path,name←⎕IO⊃LA                ⍝ next name to search
          ç93←⎕EX'ç92'                       ⍝ better be safe than sorry
          pcr←0                              ⍝ proper ⎕CR?
          :If var←onfile
              →(≥/1↓LA←LA+inc)⍴çenD          ⍝ over with the file?
              id←'Cpt ',(⍕1↓2↑LA),': '       ⍝ remember path for display
              ç92←⎕FREAD 2↑LA                ⍝ fetch the object to search from file
          :ElseIf var←∨/9 2∊⎕NC id
              ç92←⍎id                        ⍝ should not be shared!
          :Else                              ⍝ it is a function (or else, even undefined!)
              pcr←1≡≡ç92←⎕CR id              ⍝ we may have a special format for fns
          :EndIf
     
     ⍝ We have the object.
     ⍝ If it's character or numeric (and wanted): we can search AND replace.
     ⍝ If it's enclosed we only show if it's complex. If it's a ns we recurse if desired.
          :If 'P'=Typ←id TYPE ç92
         ⍝ Types are: Scripted, Char, Num, Packed (NS), Enclosed and Invalid (like Null)
              t←çR0
              :If ⎕THIS≢CurSpace ⍝ avoid infinite loops
                  t←çMrG ç92 SS 1
              :EndIf
              çRz←çRz,t
              action←0 0 0 0
          :Else  ⍝ to replace we must either have a char matrix or a list of char vectors; none must contain CR/NLs
              :If SALTed←Typ='S'
                  :If SALTed←9∊ç92.⎕NC'SALT_Data' ⋄ SALTed←{⍵.({⍵(⍎¨⍵)}⎕NL-2)}ç92.SALT_Data ⋄ :EndIf
                  ç92←⎕SRC ç92
              :EndIf ⍝ change classes' code into a string
              action←btc[btc[;0]⍳Typ;]
              t1←∨/((⍴⍴ç92)↓0 2 1)=ç90←|≡ç92
              :If t1∧2=ç90 ⍝ enclosed, could be anything. The only way to tell:
                  :If 1∨.<∊⍴¨⍴¨ç92           ⍝ if  all vectors or scalars
                  :OrIf ~∧/(10|,⎕DR¨ç92)∊0 2 ⍝ and all character, then all is well
                      t1←0                   ⍝ otherwise can't replace
                  :EndIf
              :EndIf
              ⍝ There are more constraints so we can replace
              :If t1∧2∊ç90,⍴⍴ç92 ⍝ must be deep 2 OR matrix
              :AndIf t1←~∨/(NL,CR)∊⍕ç92  ⍝ no such chars allowed
              :AndIf 2∊ç90               ⍝ enclosures only
                  ç92←1↓∊CR,¨ç92
              :EndIf
        ⍝ One more thing: if in REPLACE mode, a fn can only be dealt with if it's not a ref
              :If REPLACE∧pcr ⋄ t1←name≡(⎕NS'').⎕FX ç92 ⋄ :EndIf ⍝ is it the same name?
              action[3]∧←t1 ⍝ still replace?
          :EndIf
     
         ⍝ We determine what to do: skip this item, replace, add to result
          ç93←0 1 5⍴⍴out←'' ⍝ default value
          :If action[2]∧REPLACE≤action[3] ⍝ don't search if REPLACE and not allowed to do so
             ⍝ Search & count: <SEARCH> returns an INT table of pos, len, hit#, line #, char #
              :If 0<ç98←''⍴⍴ç93←ç91 GETSTRINGS(Typ,var↓'F')SEARCH ç91←1↓(,⌽∨\⌽' '≠ç91)/,ç91←CR,⍕ç92
                  ç94←⍴ç92
                  :If action[3] ⍝ replace
                      (ç93 ç91)←ç93 REPL ç91
                      ç92←ç94 BOX ç91
                      :If onfile
                          (ç94 BOX ç91)⎕FREPLACE 2↑LA
                      :ElseIf var
                          t1←ç94 BOX ç91
                          :If ∨/ç90←'SE'=action[0] ⋄ t1←1↓¨(1,t1∊CR)⊂';',t1 ⋄ :EndIf
                          t←⍎id,'←',(ç90[0]/'0 CurSpace.⎕fix '),'t1'
                          :If 0≢SALTed             ⍝ restore SALT tag
                              'SALT_Data't.⎕NS''
                              t.SALT_Data.{⍎(⍕⎕IO⊃⍵),'←⎕io⊃⌽⍵'}SALTed
                          :EndIf
                      :Else   ⍝ it is a function. If it fails we need to adjust the search results:
                          ç93←(~'N'∊'x'TYPE CurSpace.⎕FX ç92)⌿ç93
                      :EndIf
                  :EndIf
     
                  ç90←REPLACE≤action[3] ⋄ ç93←ç90⌿ç93 ⍝ Effective replacement?
     
                  :If SHOW∧ç90
                     ⍝ Display header
                      ç90←3 6⍴'    ∇ ',((6×¯1*Typ='S')↑'  VAR Space '),'Rep ∇ '
                      ç94←'∇'∊7⍴ç91 ⋄ ç90←ç90[var×2*ç94;],id
                      out←CR,CR,ç90,1⌽') (',(⍕ç98),' found',(0∊⍴ç93)/' but no replacement made'
                      ç91←⎕FMT ç91 ⍝0\(ç90←NL≠ç91)/ç91
                      ç90←((ç90⍳ç90)=⍳⍴ç90)/ç90←ç93[;0;3]
                      :While 0<⍴ç90
                         ⍝ Display lines found here
                          ç95←(~ç94)/(5⌈⍴ç95)↑ç95←'[',(⍕ç97←1↑ç90),'] '
                          out,←,CR,ç98←ç95,RTB,ç91[ç97;]
                         ⍝ for patterns we display the whole match
                          t←ç93[;0;3]=ç97 ⋄ ç92←(t/ç93[;0;4])PIO(t/ç93[;0;1])*Pattern
                          t←(1+⍴ç98)⍴' ' ⋄ t[(⍳⍴t)∩ç92+⍴ç95]←'∧' ⋄ out,←,CR,RTB t
                          ç90←1↓ç90
                      :EndWhile
                      :If RETURN≢1 ⋄ ⎕←out ⋄ :EndIf ⍝ gather result or display now
                  :EndIf
              :EndIf
          :EndIf
          :If action[1]
              çRz←çRz çcAr id ç93 out
          :EndIf
          LA←(~onfile)↓LA
      :EndWhile
     
     çenD: :If (RA≢1)∧REPLACE
          çRz←(⍕1↓⍴çRz),' replacements made.'
      :EndIf
      ⎕EX lac↓'ç'
    ∇

    isChar←{∨/0 2∊10|⎕DR 1/⍵}

    ∇ m←s BOX v;b;d;l;⎕IO
     ⍝    <BOX> fn: this fn serves many purposes. It basically ⎕BOX the right argument.
     ⍝     If the right arg is NOT already a matrix then:
     ⍝     If called with a 0   left arg it exits AFTER boxing, using the 1st element as delimiter.
     ⍝     If called with a 1=⍴ left arg it exits BEFORE boxing, returning its right arg.
     ⍝     If called with a 2=⍴ left arg it boxes the right arg using CR then tries to match the original shape.
      ⎕IO←1 ⋄ →(2∊⍴m←v)⍴0 ⋄ →(1∊⍴s)⍴0 ⋄ m←⍬⍴v ⋄ →(s≡⍬)⍴0 ⋄ d←1↑v←((2∊⍴s)⍴CR),v
      l←1↓l-1+¯1⌽l←(~b)/⍳⍴b←d≠m←v,d ⋄ m←(⍴b)⍴(,b←l∘.≥⍳⌈/0,b←l)\b/m
      →(s≡0)⍴0 ⋄ v←s⌈⍴m←(0,-l⊥l←' '∧.=m)↓m ⋄ m←v↑m
    ∇

    ∇ Clean
      ⎕EX'ç'⎕NL⍳10
    ∇

    ∇ R←bo HIDE M;L;C;T;B;A;t;p;apl;x;msk;q
⍝ Find search mask. M is text string. bo is keep body mask.
⍝ R=1 means valid search position.
      L←1,1↓msk←M∊NL CR ⍝ find CRs and line starts
      R←msk⍲2|PATTERN   ⍝ final mask excludes NL if multiline
      :If APL∧'∇'∊7↑M ⍝ is this a fn?
          R←R∧¯1⌽L PartOS M∊'∇]' ⍝ then account for decorators
      :EndIf
⍝ Find Body, Comments and Text masks
      →0↓⍨∨/x←3↑EXCLUDE ⍝ any code exclusions?
⍝ Separate Body, Comments and Text masks
      B←C⍱T←T>C←L PartOS(M∊'⍝')>T←T∨L PartUS T←q←M=''''
      :If APL ⍝ do we have to take executable strings into account?
    ⍝ This is the pattern to look for for executable strings:
          apl←'(?i:⎕(EX|NC|CR|VR|NR|OR|AT|SIZE|NS|CS|ED|PATH|LOCK|TRAP←[^''⍝⋄\n\r]*''[ESNC]'')|⍎[∧''⍝\n\r⋄]*) *''((''''|[^''\r\n])*''[^⍝⋄\n\r]*)'
    ⍝ We extract any text to further search into
          t←çFind apl M 0 R(0 0)
      :AndIf 0<⍴p←t[;2;0]PIO t[;2;1] ⋄ msk[p]←1
    ⍝ Remove double quotes; careful not to remove too many
          msk←msk>q>q PartUS⍨q>¯1⌽q
          A←msk\1 HIDE msk/M
          B←C⍱T←T>A
      :EndIf
      :If bo ⋄ R←R∧B
      :Else ⋄ R←R>↑∨/x/B T C
      :EndIf
    ∇

    ∇ r←LowerCase r;b
     ⍝ Also works for Underbarred letters
      →(∨/b←r∊1 0↓Uset)↓0 ⋄ b←b/⍳⍴b ⋄ r[b]←Uset[0;27|(,1 0↓Uset)⍳r[b]]
    ∇
    ∇ z←a PartAS b
    ⍝ Partitioned ∧\
      z←~≠\a\z≠¯1↓1,z←(a←a≥b)/b
    ∇

    ∇ R←S PartOS P;T
     ⍝ Partitioned ∨\
      S←S∨P ⋄ R←≠\S\T≠¯1↓0,T←S/P
    ∇

    ∇ r←s PartUS p;t
     ⍝ Partitionned ≠\ (unequal scan)
      r←≠\p≠s\t≠¯1↓0,t←s/≠\¯1↓0,p
    ∇

    ∇ (ng pattern)←caret ShortCuts pattern;sl;e;good
    ⍝ Replace ⍺ and ⍵ by APL patterns
      sl←'\'≠¯1↓'x',pattern
      :If caret
          e←(⍴pattern)↑'∧'=1↑pattern ⍝ BOL anchor
          ((e∨¯1⌽sl∧'[∧'⍷pattern)/pattern)←⎕AV[235] ⍝ map ∧ to ASCII caret if no translation done
      :EndIf
      good←e/⍳⍴e←sl∧'⍵'=pattern ⍝ numbers before ⍺ tokens
     ⍝ For APL code we accept ⍺ & ⍵ as shortcuts for 'object name' and 'number'
      pattern[e/⍳⍴e←sl∧'⍺'=pattern]←⊂AP
      pattern[good]←⊂OP
      pattern←⊃,/pattern,⊂'' ⍝ cover empty case
      pattern←(~'\⍺'⍷e)/e←(~'\⍵'⍷e)/e←pattern
      ng←1++/('('=pattern)>(¯1↓0,pattern='\')∨1↓(pattern='?'),0
    ∇

     ⍝ The 3 next functions are used in conjunction with <fr> to find where the pattern is found.
     ⍝ They return an int list indicating the start & length of each match.
     ⍝ Each subexpression specified is shown for each match (3D result).
     ⍝ They are ⎕io independent.
     ⍝ The pattern may not overlap.
     ⍝ ls/rs are left/right "syntactic" op. They work with 'notalpha' (see below).
     ⍝ 'mask' is where matches must occur entirely.

    ∇ r←{options}PATMATCHrx arg;pattern;string;notalpha;syn;mask;⎕IO;⎕ML;⎕WX;hits;z;good
      arg←arg,(⍴arg)↓0 0 0 '' 0
      (pattern string notalpha mask syn)←,¨arg
      mask←mask,((0∊⍴mask)×⍴string)⍴1
      options←(options=0 0 1)/('DotAll' 1)('Mode' 'D')('Mode' 'M')
      (r pattern)←1 ShortCuts pattern
      r←0⍴⊂(r,2)⍴0 ⍝ default result
      hits←pattern ⎕S{⍵.(Offsets,[0.1]Lengths)}⎕OPT options+string
⍝ Screen out masked positions
      :While 0<⍴hits
          z←,0⊃hits ⋄ good←1
     ⍝ notalpha is where no name forming chars are
          :If ∨/syn
              good←notalpha[0 1++\2⍴z]∧.≥syn
          :EndIf
          good←good>0∊mask[z[0]+⍳z[1]]
          r←r,good⍴hits ⋄ hits←1↓hits
          mask[⍳1⌈z[1]]←0 ⍝ do not allow overlap
      :EndWhile
      r←⊃r
    ∇

    ∇ r←{options}PATMATCHdll arg;pattern;string;notalpha;syn;mask;⎕IO;good;code;e;sl;ptr;regcomp;regerror;regexec;regfree;sxi;z;C;NULL;⎕ML;⎕WX;uni
      arg←arg,(⍴arg)↓0 0 0 '' 0
      (pattern string notalpha mask syn)←,¨arg
      mask←mask,((0∊⍴mask)×⍴string)⍴1
     
     ⍝ This will not work with Unicode chars
      :If uni←80=⎕DR''
         ⍝ (pattern string)←{⎕UCS'UTF-8'⎕UCS ⍵}¨pattern string
          'Unable to deal with Unicode chars'⎕SIGNAL 11 if 1∊255<⎕UCS pattern,string
      :EndIf
     
     ⍝ Define the regex functions (T used differently in Unicode & Classic):
      'I pcre|regexec <C[] <0T I =I[] I'⎕NA⍨'regexec'
      'I pcre|regcomp =C[] <0T I'⎕NA⍨'regcomp'
      'I pcre|regerror I <C[] =0T[] I'⎕NA⍨'regerror'
      '  pcre|regfree <C[]'⎕NA⍨'regfree'
     
      ⎕WX←3 ⋄ ⎕ML←⎕IO←0 ⋄ C←32⍴' '
      (sxi pattern)←('0'∧.=2 ⎕NQ'.' 'getenvironment' 'MAPCHARS')ShortCuts pattern
     
     ⍝ allow to show all subexpressions
      z←(2×sxi)⍴¯1
      r←0⍴⊂(sxi,2)⍴0 ⍝ default result
     
     ⍝ compile the expression into C, 0=singleline, 1=multiline
      code←regcomp C pattern(16×~options)
      (code C)←2↑code
      →code⍴err
     
     ⍝ execute the expression until nothing found
      ptr←mask⍳1   ⍝ non-overlap pointer
      :While ptr<⍴string
          z[]←-good←1
          (code z)←2↑regexec C(ptr↓string)sxi z 0
          →code⍴end
         ⍝ notalpha is where no name forming chars are
          :If ∨/syn
              good←notalpha[ptr+0 1+2⍴z]∧.≥syn
          :EndIf
          good←good>0∊mask[ptr+z[0]+⍳-/z[1 0]]
          r←r,good/⊂(sxi,2)⍴z+ptr×z≥0
          mask[ptr+⍳1⌈z[1]]←0 ⍝ do not allow overlap
          ptr←z[1]⌈mask⍳1
      :EndWhile
     
     end:
      r[;;1]←|-/r←↑r ⍝ 2nd element is the length
      z←regfree,⊂C
      →0
     ⍝
     err:z←256⍴' '
      r←regerror code C z 255 ⍝ we expect no more than a 255 chars msg
      z←regfree,⊂C
      (r z)←r
      e←82 80[uni]⎕DR r↑z ⍝ the error message
      e ⎕SIGNAL 11
    ∇

    ∇ r←{options}PATMATCHdnet arg;⎕USING;CR;LF;syn;mo;np;len;ind;pattern;string;notalpha;mask;ls;rs;⎕IO;good;code;e;sl;ptr;regcomp;regerror;regexec;regfree;sxi;z;C;NULL;⎕ML;⎕WX
      arg←arg,(⍴arg)↓0 0 0 '' 0
      (pattern string notalpha mask syn)←,¨arg
      mask←mask,((0∊⍴mask)×⍴string)⍴1
     
      ⎕WX←3 ⋄ ⎕ML←⎕IO←0
      (CR LF)←⎕AV[3 2]
      (r pattern)←1 ShortCuts pattern
      r←0⍴⊂(r,2)⍴0 ⍝ default result
     
     ⍝ Determine the options (0=singleline, 1=multiline)
      ⎕USING←'System.Text.RegularExpressions,system.dll'
      options←options⊃RegexOptions.(Singleline Multiline)
     
⍝ For multiline to work under .Net we need NLs
      ((string=CR)/string)←LF
     
      mo←Regex.Match string pattern options
     
     ⍝ Execute the expression until nothing found
      :While good←mo.Success
          (len ind)←mo.(Length Index)
          :If ∨/syn
     ⍝ notalpha is where no name forming chars are (padded 1 left and right to avoid INDEX error)
     ⍝ This supplements the ability of the engine to detect word characters with \w as Dyalog
     ⍝ also uses other characters like Á. 'notalpha' can indicate where those are.
              good←notalpha[ind+0,len+1]∧.≥syn
          :EndIf
          :If good>0∊(⊂ind+⍳len)⌷mask,0
              np←mo.Groups.Count
              r←r,⊂↑mo.Groups[⍳np].(Index Length)
          :EndIf
          mo←mo.NextMatch
      :EndWhile
     
     end:
      r←↑r ⍝ disclose results
    ∇

    if←/⍨

    ∇ r←a PIO b
     ⍝ Progressive IOTA
      r←(b/a)+(⍳+/b)-b/¯1↓0,+\b
    ∇

    ∇ REDEFINE ns;t;alls
      alls←(ns.⎕NL-2)~'Arguments' 'SwD' ⍝ all switches names
      ⍎t,'←ns.(',(t←⍕alls),')'          ⍝ redefine all vars locally
      Strings←ns.Arguments              ⍝ the strings to search
    ∇

    ∇ r←p REPL t;k;b;Tl;Fp;Fl;∆;rstr
      ⍝ Replace
      Tl←⊃,/⍴¨rstr←REPLX t p Rstr ⋄ ∆←Tl-Fl←p[;0;1]
      k←(Fp,⍴t)-0,Fl+Fp←p[;0;0]
      b←(2×⍴k)⍴1 0
      t←((,k,[0.1]Fl,0)/b)/t ⋄ t←(b←(,k,[0.1]Tl,0)/b)\t
      t[(~b)/⍳⍴b]←⊃,/rstr ⋄ Fl←p[;0;0]←p[;0;0]+¯1↓0,+\∆ ⋄ p[;0;1]←Tl
      Fp←b/⍳⍴b←CR=t ⋄ p[;0;4]←Fl-b+×b←(0,Fp)[p[;0;3]←+/Fl∘.>Fp] ⋄ r←p t
    ∇

    ∇ r←REPLX(string pos pats);N;new;strs;b;nosl;ip;i
      ⍝ Replace RegeX patterns fn
      r←pats[ip←pos[;0;2]]
      →0↓⍨REPLACE∧PATTERN>0
      r←⍳0 ⋄ N←(⍴pos)[1]↑'0123456789'
      :For i :In ⍳⍴ip
          nosl←~1⌽b←(new∊N)∧¯1⌽'\'=new←⊃pats[ip[i]]
          :If ∨/b
              strs←pos[i;;1]↑¨pos[i;;0]↓¨⊂string
              (b/new)←strs[N⍳b/new] ⋄ new←⊃,/nosl/new
          :EndIf
          r←r,⊂new
      :EndFor
    ∇

    RTB←{⍵↓⍨-⊥⍨' '=⍵} ⍝ Remove Trailing Blanks

    ∇ r←typ SEARCH t;h;m;l;n;S;i;s;w;ll1;nl;R;r0;⎕TRAP;syn;L1;cr;lim
      r←r0←0 1 5⍴0
     ⍝ Are we excluding BOTH locals and globals?
      →0/⍨∧/L1←¯2↑EXCLUDE∧(∨/'FS'∊typ)∨'      ∇'≡7↑t ⍝ for fns only
      →(⍴t)↓i←0 ⋄ nl←cr/⍳⍴cr←t∊CR,NL
      n←⍴Strings ⋄ r←0 1 3⍴0
      :If ∨/L1 ⍝ do we need to exclude/include those in fn headers?
          ll1←(⍴t)↑((1↑nl)⌊t⍳'⍝')⍴1 ⋄ lim←2 1⍴0,⍴t
      :AndIf 'S'∊typ ⍝ account for scripted form here
          s←s⍱cr PartOS(cr PartUS t='⍝')>s←cr PartUS t=''''
          m←s∧t∊'{}' ⋄ w←0≠+\¯1*'}'=m/t ⋄ s>←¯1⌽≠\m\w≠¯1↓0,w ⍝ mask out D-fns
         ⍝ Any line that starts with a ∇ is now a Tradfn header or footer
          ll1←s∧cr PartOS(¯1⌽cr PartAS cr∨' '=t)∧'∇'=t
          lim←⍉((0.5×⍴m),2)⍴m←nl/⍨1∊¨cr⊂ll1 ⍝ each fn's line limits
      :EndIf
      m←0 HIDE t
      ⎕TRAP←DEBUG↓⊂0 'C' '⎕DM⋄→⍴r←r0'
      :If syn←S←∨/SYNTACTIC ⋄ S←~0,(t∊Cset),0 ⋄ :EndIf
      :If INSENSITIVE ⋄ t←LowerCase t :EndIf
     ⍝ OK, we're set: search each string in turn in the present code
     l1:→(w←⍴s←i⊃Strings)↓e
      :If 0<PATTERN
          R←çFind s t S m SYNTACTIC ⋄ h←R[;0;0] ⋄ w←R[;0;1]
      :Else
          h←{⍵/⍳⍴⍵}s⍷t
          :If syn
              h←(S[h∘.+0,1+w]∧.≥SYNTACTIC)/h
          :EndIf
          h←(∧/m[h∘.+⍳w])/h ⍝ all valid?
      :EndIf
     ⍝ If any hit in the first line we skip this item if so desired
      :If 1∊L1
       ⍝ Find which fn each hit belongs to
          l←ll1[h]/s←(1+⍳1↓⍴lim)+.×≠⌿lim∘.<h   ⍝ which fns hit in header
          h←(L1[1]=s∊l)/h
      :EndIf
      →(⍴h)↓e
      :If REPLACE>PATTERN ⍝ in replace mode update mask as we go along
          :While ~∧/s←≤\h≥¯1↓0,w+h ⋄ h←s/h ⋄ :EndWhile ⍝ remove overlaps
      :EndIf
      s←⍴h ⋄ w←s⍴w
      :If REPLACE ⋄ m[h PIO w]←0 ⋄ :EndIf  ⍝ update mask (w may be variable)
      :If 0=PATTERN ⋄ R←(s,1 2)⍴h,[0.1]w   ⍝ add length
      :ElseIf 0=i ⋄ r←R,i ⋄ →e ⍝ 1st time set r
      :EndIf
     l8:r←r{a←a⌈m←0,1↓(a←⍴⍺)⌈w←⍴⍵ ⋄ (a↑⍺)⍪(m⌈w)↑⍵}R,i
     e:→(n>i←i+h←R←1+REPLACE)/l1 ⋄ r←r[⍋r;;] ⍝ order result
      l←+⌿nl∘.<s←r[;;0] ⋄ r←(r,l),s-w+×w←(0,nl)[l] ⍝ lines #,  position in it
      r[;;2]←r[;;2]÷R ⍝ adjust string #
    ∇

    ∇ r←nam TYPE val;t;s
     ⍝ Return a 1 letter defining the type of the value
      r←'SI'⌷⍨{0::1 ⋄ 0{⍺}⎕SRC ⍵}val   ⍝ scripted? I=Invalid
      →0/⍨r≠'I'
      →(9∊⎕NC nam)⍴l10×RECURSIVE ⍝ ←global
      t←≡val ⋄ →(s←+/⍴⍴val)⍴ç ⋄ →(0≡t)↓ç ⋄ →ç×'[Null]'≢⍕val
     l10:→(∨/'( '∊1⍴⍕val)⍴0
      :Trap 6 ⋄ r←(⍎nam).⎕WG'type' ⋄ :Else ⋄ r←(⍎nam).Type ⋄ :EndTrap
      r←'SSPII'['Cla' 'Int' 'Nam' 'Net'⍳⊂3↑r] ⋄ →0
     ç:r←'E' ⋄ →((s=0)∧1≡t)/0 ⍝ ⎕OR?
      r←'ECCN'[t←6 0 2⍳10|⎕DR 1/val] ⋄ →t↓0 ⋄ →(3>⍴⍴val)⍴0 ⋄ r←'E'
    ∇

:Endnamespace